#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2005-2009  Serge Noiraud
# Copyright (C) 2007-2009  Brian G. Matherly
# Copyright (C) 2010       Peter Landgren
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011       Adam Stein <adam@csh.rit.edu>
# Copyright (C) 2012,2017  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
ODFDoc : used to generate Open Office Document
"""

# -------------------------------------------------------------------------
#
# Standard Python Modules
#
# -------------------------------------------------------------------------
import os
from hashlib import md5
import zipfile
import time
from io import StringIO
from math import cos, sin, radians
from xml.sax.saxutils import escape
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.plug.docgen import (
    BaseDoc,
    TextDoc,
    DrawDoc,
    graphicstyle,
    FONT_SANS_SERIF,
    SOLID,
    PAPER_PORTRAIT,
    INDEX_TYPE_TOC,
    PARA_ALIGN_CENTER,
    PARA_ALIGN_LEFT,
    INDEX_TYPE_ALP,
    PARA_ALIGN_RIGHT,
    URL_PATTERN,
    LOCAL_HYPERLINK,
    LOCAL_TARGET,
)
from gramps.gen.plug.docgen.fontscale import string_width
from gramps.plugins.lib.libodfbackend import OdfBackend
from gramps.gen.const import PROGRAM_NAME, URL_HOMEPAGE
from gramps.version import VERSION
from gramps.gen.plug.report import utils
from gramps.gen.utils.image import (
    image_size,
    image_dpi,
    image_actual_size,
    crop_percentage_to_subpixel,
)
from gramps.gen.errors import ReportError

# -------------------------------------------------------------------------
#
# internationalization
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

APP_TYPE = "application/vnd.oasis.opendocument.text"

ESC_MAP = {
    "\x1a": "",
    "\x0c": "",
    "\n": "<text:line-break/>",
    "\t": "<text:tab />",
}

# -------------------------------------------------------------------------
#
# regexp for Styled Notes ...
#
# -------------------------------------------------------------------------
# Hyphen is added because it is used to replace spaces in the font name
NEW_STYLE = re.compile('style-name="([a-zA-Z0-9]*)__([#a-zA-Z0-9 -]*)__">')

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

_XMLNS = """\
xmlns:office="%(urn)soffice:1.0"
xmlns:style="%(urn)sstyle:1.0"
xmlns:text="%(urn)stext:1.0"
xmlns:table="%(urn)stable:1.0"
xmlns:draw="%(urn)sdrawing:1.0"
xmlns:fo="%(urn)sxsl-fo-compatible:1.0"
xmlns:xlink="http://www.w3.org/1999/xlink"
xmlns:dc="http://purl.org/dc/elements/1.1/"
xmlns:meta="%(urn)smeta:1.0"
xmlns:number="%(urn)sdatastyle:1.0"
xmlns:svg="%(urn)ssvg-compatible:1.0"
xmlns:chart="%(urn)schart:1.0"
xmlns:dr3d="%(urn)sdr3d:1.0"
xmlns:math="http://www.w3.org/1998/Math/MathML"
xmlns:form="%(urn)sform:1.0"
xmlns:script="%(urn)sscript:1.0"
xmlns:dom="http://www.w3.org/2001/xml-events"
xmlns:xforms="http://www.w3.org/2002/xforms"
""" % {
    "urn": "urn:oasis:names:tc:opendocument:xmlns:"
}

_FONTS = """\
<style:font-face style:name="Courier"
    svg:font-family="Courier"
    style:font-family-generic="modern"
    style:font-pitch="fixed"/>

<style:font-face style:name="Times New Roman"
    svg:font-family="&apos;Times New Roman&apos;"
    style:font-family-generic="roman"
    style:font-pitch="variable"/>

<style:font-face style:name="Arial"
    svg:font-family="Arial"
    style:font-family-generic="swiss"
    style:font-pitch="variable"/>

"""

_META_XML = """\
<?xml version="1.0" encoding="UTF-8"?>
<office:document-meta
    xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
    office:version="1.0">
<office:meta>
<meta:generator>
    %(generator)s
</meta:generator>
<dc:title>
</dc:title>
<dc:subject>
</dc:subject>
<dc:description>
</dc:description>
<meta:initial-creator>
    %(creator)s
</meta:initial-creator>
<meta:creation-date>
    %(date)s
</meta:creation-date>
<dc:creator>
    %(creator)s
</dc:creator>
<dc:date>
    %(date)s
</dc:date>
<meta:print-date>0-00-00T00:00:00</meta:print-date>
<dc:language>
    %(lang)s
</dc:language>
<meta:editing-cycles>1</meta:editing-cycles>
<meta:editing-duration>PT0S</meta:editing-duration>
<meta:user-defined
    meta:name="Genealogical Research and Analysis Management Programming System">
    %(gramps_home_url)s
</meta:user-defined>
<meta:user-defined meta:name="Info 1"/>
<meta:user-defined meta:name="Info 2"/>
<meta:user-defined meta:name="Info 3"/>
</office:meta>
</office:document-meta>
"""

_STYLES = """\
<style:default-style
    style:family="graphic">
    <style:graphic-properties
        draw:shadow-offset-x="0.3cm"
        draw:shadow-offset-y="0.3cm"
        draw:start-line-spacing-horizontal="0.283cm"
        draw:start-line-spacing-vertical="0.283cm"
        draw:end-line-spacing-horizontal="0.283cm"
        draw:end-line-spacing-vertical="0.283cm"
        style:flow-with-text="true"/>
    <style:paragraph-properties
        style:text-autospace="ideograph-alpha"
        style:line-break="strict"
        style:writing-mode="lr-tb"
        style:font-independent-line-spacing="false">
        <style:tab-stops/>
    </style:paragraph-properties>
    <style:text-properties
        style:use-window-font-color="true"
        fo:font-size="12pt"
        style:font-size-asian="12pt"
        style:language-asian="none"
        style:country-asian="none"
        style:font-size-complex="12pt"
        style:language-complex="none"
        style:country-complex="none"/>
</style:default-style>

<style:default-style
     style:family="paragraph">
     <style:paragraph-properties
         style:text-autospace="ideograph-alpha"
         style:punctuation-wrap="hanging"
         style:line-break="strict"
         style:tab-stop-distance="2.205cm"
         style:writing-mode="page"/>
    <style:text-properties
        style:font-name="Times New Roman"
        fo:font-size="12pt"
        style:font-name-asian="Times New Roman"
        style:font-size-asian="12pt"
        style:font-name-complex="Times New Roman"
        style:font-size-complex="12pt"
        style:tab-stop-distance="2.205cm"/>
</style:default-style>

<style:default-style
     style:family="table">
     <style:table-properties
        table:border-model="separating"/>
</style:default-style>

<style:default-style
    style:family="table-row">
    <style:table-row-properties
        fo:keep-together="auto"/>
</style:default-style>

<style:style style:name="Standard"
    style:family="paragraph" style:class="text"/>

<style:style style:name="photo"
    style:family="graphic">
    <style:graphic-properties
        text:anchor-type="paragraph"
        svg:x="0cm" svg:y="0cm" style:wrap="none"
        style:vertical-pos="top"
        style:vertical-rel="paragraph-content"
        style:horizontal-pos="center"
        style:horizontal-rel="paragraph-content"/>
</style:style>
"""

_AUTOMATIC_STYLES = """\
<style:style style:name="docgen_page_break"
    style:family="paragraph"
    style:parent-style-name="Standard">
    <style:paragraph-properties
        fo:break-before="page"/>
</style:style>

<style:style style:name="GSuper"
    style:family="text">
    <style:text-properties
        style:text-position="super 58%"/>
</style:style>

<style:style style:name="GSub"
    style:family="text">
    <style:text-properties
        style:text-position="sub 58%"/>
</style:style>

<style:style style:name="GStrikethrough"
    style:family="text">
    <style:text-properties
        style:text-line-through-style="solid"/>
</style:style>

<style:style style:name="GRAMPS-preformat"
    style:family="text">
    <style:text-properties
        style:font-name="Courier"/>
</style:style>

"""

_CLEAR_STYLE = """\
<style:style style:name="clear"
    style:family="graphic">\n
    <style:graphic-properties draw:stroke="none"
        draw:fill="none" draw:shadow="hidden"
        style:run-through="foreground"
        style:vertical-pos="from-top"
        style:vertical-rel="paragraph"
        style:horizontal-pos="from-left"
        style:horizontal-rel="paragraph"
        draw:wrap-influence-on-position="once-concurrent"
        style:flow-with-text="false"/>
</style:style>\n
"""
_OTHER_STYLES = """\
<style:style style:name="Tbold"
    style:family="text">\n
    <style:text-properties fo:font-weight="bold"/>\n
</style:style>\n

<style:style style:name="Titalic"
    style:family="text">\n
    <style:text-properties fo:font-style="italic"/>\n
</style:style>\n

<style:style style:name="Tunderline"
    style:family="text">\n
    <style:text-properties
        style:text-underline-style="solid"
        style:text-underline-width="auto"
        style:text-underline-color="font-color"/>
</style:style>\n

<style:style style:name="Left"
    style:family="graphic"
     style:parent-style-name="photo">
    <style:graphic-properties
        style:run-through="foreground"
         style:wrap="dynamic"
         style:number-wrapped-paragraphs="no-limit"
         style:wrap-contour="false"
         style:vertical-pos="from-top"
         style:vertical-rel="paragraph-content"
         style:horizontal-pos="left"
         style:horizontal-rel="paragraph-content"
         style:mirror="none" fo:clip="rect(0cm 0cm 0cm 0cm)"
         draw:luminance="0%" draw:contrast="0" draw:red="0%"
         draw:green="0%" draw:blue="0%" draw:gamma="1"
         draw:color-inversion="false"
         draw:transparency="-100%"
         draw:color-mode="standard"/>
</style:style>\n

<style:style style:name="Right"
    style:family="graphic"
     style:parent-style-name="photo">
    <style:graphic-properties
        style:run-through="foreground"
         style:wrap="dynamic"
         style:number-wrapped-paragraphs="no-limit"
         style:wrap-contour="false"
         style:vertical-pos="from-top"
         style:vertical-rel="paragraph-content"
         style:horizontal-pos="right"
         style:horizontal-rel="paragraph-content"
         style:mirror="none" fo:clip="rect(0cm 0cm 0cm 0cm)"
         draw:luminance="0%" draw:contrast="0" draw:red="0%"
         draw:green="0%" draw:blue="0%" draw:gamma="1"
         draw:color-inversion="false"
         draw:transparency="-100%"
         draw:color-mode="standard"/>
</style:style>\n

<style:style style:name="Single"
    style:family="graphic"
     style:parent-style-name="Graphics">
    <style:graphic-properties
        style:vertical-pos="from-top"
         style:mirror="none" fo:clip="rect(0cm 0cm 0cm 0cm)"
         draw:luminance="0%" draw:contrast="0" draw:red="0%"
         draw:green="0%" draw:blue="0%" draw:gamma="1"
         draw:color-inversion="false"
         draw:transparency="-100%"
         draw:color-mode="standard"/>
</style:style>\n

<style:style style:name="Row" style:family="graphic"
     style:parent-style-name="Graphics">
    <style:graphic-properties
        style:vertical-pos="from-top"
         style:vertical-rel="paragraph"
         style:horizontal-pos="from-left"
         style:horizontal-rel="paragraph"
         style:mirror="none" fo:clip="rect(0cm 0cm 0cm 0cm)"
         draw:luminance="0%" draw:contrast="0" draw:red="0%"
         draw:green="0%" draw:blue="0%" draw:gamma="1"
         draw:color-inversion="false"
         draw:transparency="-100%"
         draw:color-mode="standard"/>
</style:style>\n
"""

_SHEADER_FOOTER = """\
<style:style style:name="S-Header"
    style:family="paragraph"
    style:parent-style-name="Standard">
    <style:paragraph-properties fo:text-align="center"
        style:justify-single-word="false"/>
</style:style>\n
<style:style style:name="S-Footer"
    style:family="paragraph"
    style:parent-style-name="Header">
    <style:paragraph-properties fo:text-align="center"
        style:justify-single-word="false"/>
</style:style>\n
"""

_CLICKABLE = r"""<text:a xlink:type="simple" xlink:href="\1">\1</text:a>"""


# -------------------------------------------------------------------------
#
# ODFDoc
#
# -------------------------------------------------------------------------
class ODFDoc(BaseDoc, TextDoc, DrawDoc):
    """
    The ODF document class
    """

    def __init__(self, styles, ftype, uistate=None):
        """
        Class init
        """
        BaseDoc.__init__(self, styles, ftype, uistate=uistate)
        self.media_list = []
        self.init_called = False
        self.index_title = None
        self.toc_title = None
        self.cntnt = None
        self.cntnt1 = None
        self.cntnt2 = None
        self.cntntx = None
        self.sfile = None
        self.mimetype = None
        self.meta = None
        self.mfile = None
        self.stfile = None
        self.filename = None
        self.lang = None
        self._backend = None
        self.column_order = None
        self.row_cells = None
        self.span = 0
        self.level = 0
        self.time = "0000-00-00T00:00:00"
        self.new_page = 0
        self.new_cell = 0
        self.page = 0
        self.first_page = 1
        self.stylelist_notes = []  # styles to create for styled notes.
        self.stylelist_photos = []  # styles to create for clipped images.

    def open(self, filename):
        """
        Open the new document
        """
        now = time.localtime(time.time())
        self.time = "%04d-%02d-%02dT%02d:%02d:%02d" % now[:6]

        self.filename = filename
        if not filename.endswith("odt"):
            self.filename += ".odt"

        self.filename = os.path.normpath(os.path.abspath(self.filename))
        self._backend = OdfBackend()
        self.cntnt = StringIO()
        self.cntnt1 = StringIO()
        self.cntnt2 = StringIO()

    def init(self):
        """
        Create the document header
        """

        assert not self.init_called
        self.init_called = True
        wrt = self.cntnt.write
        wrt1, wrt2 = self.cntnt1.write, self.cntnt2.write

        self.lang = glocale.lang
        self.lang = self.lang.replace("_", "-") if self.lang else "en-US"

        self.stylelist_notes = []  # styles to create depending on styled notes.
        wrt1(
            '<?xml version="1.0" encoding="UTF-8"?>\n'
            + "<office:document-content\n"
            + _XMLNS
            + 'office:version="1.0">\n'
            + "<office:scripts/>\n"
        )
        wrt1("<office:font-face-decls>\n" + _FONTS)
        wrt2(
            "</office:font-face-decls>\n"
            + "<office:automatic-styles>\n"
            + _AUTOMATIC_STYLES
        )

        styles = self.get_style_sheet()

        for style_name in styles.get_draw_style_names():
            style = styles.get_draw_style(style_name)
            wrt(
                "<style:style "
                + 'style:name="%s" ' % style_name
                + 'style:family="graphic">\n'
                "<style:graphic-properties "
            )

            if style.get_line_width():
                wrt(
                    'svg:stroke-width="%.2f" ' % (style.get_line_width() * 10)
                    + 'draw:marker-start="" '
                    'draw:marker-start-width="0.0" '
                    'draw:marker-end-width="0.0" '
                    'draw:textarea-horizontal-align="center" '
                    'draw:textarea-vertical-align="middle" '
                )
                if style.get_line_style() != SOLID:
                    # wrt('svg:fill-color="#ff0000" ')
                    wrt(
                        'draw:stroke="dash" draw:stroke-dash="gramps_%s" '
                        % style.get_dash_style_name()
                    )
                else:
                    wrt('draw:stroke="solid" ')
            else:
                wrt('draw:stroke="none" ' 'draw:stroke-color="#000000" ')

            wrt(
                'svg:fill-color="#%02x%02x%02x" ' % style.get_color()
                + 'draw:fill-color="#%02x%02x%02x" ' % style.get_fill_color()
                + 'draw:shadow="hidden" '
                'style:run-through="foreground" '
                'style:vertical-pos="from-top" '
                'style:vertical-rel="paragraph" '
                'style:horizontal-pos="from-left" '
                'style:horizontal-rel="paragraph" '
                "draw:wrap-influence-on-position="
                '"once-concurrent" '
                'style:flow-with-text="false" '
                "/>\n"
                "</style:style>\n"
            )

            wrt(
                "<style:style "
                'style:name="%s_shadow" ' % style_name + 'style:family="graphic">\n'
                "<style:graphic-properties "
                'draw:stroke="none" '
                'draw:fill="solid" '
                'draw:fill-color="#cccccc" '
                'draw:textarea-horizontal-align="center" '
                'draw:textarea-vertical-align="middle" '
                'draw:shadow="hidden" '
                'style:run-through="foreground" '
                'style:vertical-pos="from-top" '
                'style:vertical-rel="paragraph" '
                'style:horizontal-pos="from-left" '
                'style:horizontal-rel="paragraph" '
                "draw:wrap-influence-on-position="
                '"once-concurrent" '
                'style:flow-with-text="false" '
                "/>\n"
                "</style:style>\n"
            )

        # Graphic style for items with a clear background
        wrt(_CLEAR_STYLE)

        for style_name in styles.get_paragraph_style_names():
            style = styles.get_paragraph_style(style_name)

            wrt(
                '<style:style style:name="NL%s" ' % style_name
                + 'style:family="paragraph" '
                + 'style:parent-style-name="%s">\n' % style_name
                + "<style:paragraph-properties "
                + 'fo:break-before="page"/>\n'
                + "</style:style>\n"
                + '<style:style style:name="X%s" ' % style_name
                + 'style:family="paragraph"'
                + ">\n"
                + "<style:paragraph-properties "
            )

            if style.get_padding() != 0.0:
                wrt('fo:padding="%.2fcm" ' % style.get_padding())
            if style.get_header_level() > 0:
                wrt('fo:keep-with-next="auto" ')

            align = style.get_alignment()
            if align == PARA_ALIGN_LEFT:
                wrt('fo:text-align="start" ')
            elif align == PARA_ALIGN_RIGHT:
                wrt('fo:text-align="end" ')
            elif align == PARA_ALIGN_CENTER:
                wrt('fo:text-align="center" ' 'style:justify-single-word="false" ')
            else:
                wrt('fo:text-align="justify" ' 'style:justify-single-word="false" ')
            font = style.get_font()
            wrt(
                'style:font-name="%s" '
                % (
                    "Arial"
                    if font.get_type_face() == FONT_SANS_SERIF
                    else "Times New Roman"
                )
            )

            wrt(
                'fo:font-size="%.2fpt" ' % font.get_size()
                + 'style:font-size-asian="%.2fpt" ' % font.get_size()
                + 'fo:color="#%02x%02x%02x" ' % font.get_color()
            )

            if font.get_bold():
                wrt('fo:font-weight="bold" ')
            if font.get_italic():
                wrt('fo:font-style="italic" ')
            if font.get_underline():
                wrt(
                    'style:text-underline="single" '
                    'style:text-underline-color="font-color" '
                )

            wrt(
                'fo:text-indent="%.2fcm"\n' % style.get_first_indent()
                + 'fo:margin-right="%.2fcm"\n' % style.get_right_margin()
                + 'fo:margin-left="%.2fcm"\n' % style.get_left_margin()
                + 'fo:margin-top="%.2fcm"\n' % style.get_top_margin()
                + 'fo:margin-bottom="%.2fcm"\n' % style.get_bottom_margin()
                + "/>\n"
                + "</style:style>\n"
            )

            wrt(
                '<style:style style:name="F%s" ' % style_name
                + 'style:family="text">\n'
                + "<style:text-properties "
            )

            align = style.get_alignment()
            if align == PARA_ALIGN_LEFT:
                wrt('fo:text-align="start" ')
            elif align == PARA_ALIGN_RIGHT:
                wrt('fo:text-align="end" ')
            elif align == PARA_ALIGN_CENTER:
                wrt('fo:text-align="center" ' 'style:justify-single-word="false" ')

            font = style.get_font()
            wrt(
                'style:font-name="%s" '
                % (
                    "Arial"
                    if font.get_type_face() == FONT_SANS_SERIF
                    else "Times New Roman"
                )
            )

            color = font.get_color()
            wrt('fo:color="#%02x%02x%02x" ' % color)
            if font.get_bold():
                wrt('fo:font-weight="bold" ')
            if font.get_italic():
                wrt('fo:font-style="italic" ')

            wrt(
                'fo:font-size="%.2fpt" ' % font.get_size()
                + 'style:font-size-asian="%.2fpt"/> ' % font.get_size()
                + "</style:style>\n"
            )

        for style_name in styles.get_table_style_names():
            style = styles.get_table_style(style_name)
            table_width = float(self.get_usable_width())
            table_width_str = "%.2f" % table_width
            wrt(
                '<style:style style:name="%s" ' % style_name
                + 'style:family="table-properties">\n'
                "<style:table-properties-properties "
                + 'style:width="%scm" ' % table_width_str
                + "/>\n"
                + "</style:style>\n"
            )

            for col in range(0, min(style.get_columns(), 50)):
                width = table_width * float(style.get_column_width(col) / 100.0)
                width_str = "%.4f" % width
                wrt(
                    '<style:style style:name="%s.%s" '
                    % (style_name, chr(ord("A") + col))
                    + 'style:family="table-column">'
                    + "<style:table-column-properties "
                    + 'style:column-width="%scm"/>' % width_str
                    + "</style:style>\n"
                )

        for cell in styles.get_cell_style_names():
            cell_style = styles.get_cell_style(cell)
            wrt(
                '<style:style style:name="%s" ' % cell
                + 'style:family="table-cell">\n'
                + "<style:table-cell-properties"
                + ' fo:padding="%.2fcm"' % cell_style.get_padding()
            )
            wrt(
                ' fo:border-top="%s"'
                % ("0.002cm solid #000000" if cell_style.get_top_border() else "none")
            )

            wrt(
                ' fo:border-bottom="%s"'
                % (
                    "0.002cm solid #000000"
                    if cell_style.get_bottom_border()
                    else "none"
                )
            )

            wrt(
                ' fo:border-left="%s"'
                % ("0.002cm solid #000000" if cell_style.get_left_border() else "none")
            )

            wrt(
                ' fo:border-right="%s"'
                % ("0.002cm solid #000000" if cell_style.get_right_border() else "none")
            )

            wrt("/>\n" "</style:style>\n")

        wrt(_OTHER_STYLES)

        wrt(
            "</office:automatic-styles>\n"
            "<office:body>\n"
            " <office:text>\n"
            " <office:forms "
            'form:automatic-focus="false" '
            'form:apply-design-mode="false"/>\n'
        )

    def uniq(self, list_, funct=None):
        """
        We want no duplicate in the list
        """
        # order preserving
        funct = funct or (lambda x: x)
        seen = set()
        result = []
        for item in list_:
            marker = funct(item[0])
            if marker in seen:
                continue
            seen.add(marker)
            result.append(item)
        return result

    def finish_cntnt_creation(self):
        """
        We have finished the document.
        So me must integrate the new fonts and styles where they should be.
        The content.xml file is closed.
        """
        self.cntntx = StringIO()
        self.stylelist_notes = self.uniq(self.stylelist_notes)
        self.add_styled_notes_fonts()
        self.add_styled_notes_styles()
        self.add_styled_photo_styles()
        self.cntntx.write(self.cntnt1.getvalue())
        self.cntntx.write(self.cntnt2.getvalue())
        self.cntntx.write(self.cntnt.getvalue())
        self.cntnt1.close()
        self.cntnt2.close()
        self.cntnt.close()

    def close(self):
        """
        Close the document and create the odt file
        """
        self.cntnt.write(
            "</office:text>\n" "</office:body>\n" "</office:document-content>\n"
        )
        self.finish_cntnt_creation()
        self._write_styles_file()
        self._write_manifest()
        self._write_settings()
        self._write_meta_file()
        self._write_mimetype_file()
        self._write_zip()

    def add_styled_notes_fonts(self):
        """
        Add the new fonts for Styled notes in the font-face-decls section.
        """
        # Need to add new font for styled notes here.
        wrt1 = self.cntnt1.write
        for style in self.stylelist_notes:
            if style[1] == "FontFace":
                # Restore any spaces that were replaced by hyphens in
                # libodfbackend
                wrt1(
                    "<style:font-face "
                    + '    style:name="%s"\n' % style[2].replace("-", " ")
                    + '    svg:font-family="&apos;%s&apos;"\n'
                    % style[2].replace("-", " ")
                    + '    style:font-pitch="fixed"/>\n\n'
                )

    def add_styled_notes_styles(self):
        """
        Add the new styles for Styled notes in the automatic-styles section.
        """
        # Need to add new style for styled notes here.
        wrt2 = self.cntnt2.write
        for style in self.stylelist_notes:
            if style[1] == "FontSize":
                wrt2(
                    "<style:style "
                    + 'style:name="FontSize__%s__"\n' % style[2]
                    + '    style:family="text">\n'
                    + "    <style:text-properties\n"
                    + '        fo:font-size="%spt"\n' % style[2]
                    + '        style:font-size-asian="%spt"\n' % style[2]
                    + '        style:font-size-complex="%spt"/>\n' % style[2]
                    + "</style:style>\n\n"
                )

            elif style[1] == "FontColor":
                # Restore the hash at the start that was removed by
                # libodfbackend
                wrt2(
                    "<style:style "
                    + 'style:name="FontColor__%s__"\n' % style[2]
                    + '    style:family="text">\n'
                    + "    <style:text-properties\n"
                    + '        fo:color="#%s"/>\n' % style[2]
                    + "</style:style>\n\n"
                )

            elif style[1] == "FontHighlight":
                wrt2(
                    "<style:style "
                    + 'style:name="FontHighlight__%s__"\n' % style[2]
                    + '    style:family="text">\n'
                    + "    <style:text-properties\n"
                    + '        fo:background-color="#%s"/>\n' % style[2]
                    + "</style:style>\n\n"
                )

            elif style[1] == "FontFace":
                # Restore any spaces that were replaced by hyphens in
                # libodfbackend
                wrt2(
                    "<style:style "
                    + 'style:name="FontFace__%s__"\n' % style[2]
                    + '    style:family="text">\n'
                    + "    <style:text-properties\n"
                    + '        style:font-name="%s"\n' % style[2].replace("-", " ")
                    + '        style:font-pitch="variable"/>\n'
                    + "</style:style>\n\n"
                )

    def add_styled_photo_styles(self):
        """
        Add the new styles for clipped images in the automatic-styles section.
        """
        wrt2 = self.cntnt2.write
        for style in self.stylelist_photos:
            if style[0] == "Left":
                wrt2(
                    "<style:style "
                    + 'style:name="Left_%s" ' % str(style[1])
                    + 'style:family="graphic" '
                    + 'style:parent-style-name="photo">'
                    + "<style:graphic-properties "
                    + 'style:run-through="foreground" '
                    + 'style:wrap="dynamic" '
                    + 'style:number-wrapped-paragraphs="no-limit" '
                    + 'style:wrap-contour="false" '
                    + 'style:vertical-pos="from-top" '
                    + 'style:vertical-rel="paragraph-content" '
                    + 'style:horizontal-pos="left" '
                    + 'style:horizontal-rel="paragraph-content" '
                    + 'style:mirror="none" '
                    + 'fo:clip="rect(%fin %fin %fin %fin)" ' % style[1]
                    + 'draw:luminance="0%" '
                    + 'draw:contrast="0" '
                    + 'draw:red="0%" '
                    + 'draw:green="0%" '
                    + 'draw:blue="0%" '
                    + 'draw:gamma="1" '
                    + 'draw:color-inversion="false" '
                    + 'draw:transparency="-100%" '
                    + 'draw:color-mode="standard"/>'
                    + "</style:style>\n"
                )
            elif style[0] == "Right":
                wrt2(
                    "<style:style "
                    + 'style:name="Right_%s" ' % str(style[1])
                    + 'style:family="graphic" '
                    + 'style:parent-style-name="photo">'
                    + "<style:graphic-properties "
                    + 'style:run-through="foreground" '
                    + 'style:wrap="dynamic" '
                    + 'style:number-wrapped-paragraphs="no-limit" '
                    + 'style:wrap-contour="false" '
                    + 'style:vertical-pos="from-top" '
                    + 'style:vertical-rel="paragraph-content" '
                    + 'style:horizontal-pos="right" '
                    + 'style:horizontal-rel="paragraph-content" '
                    + 'style:mirror="none" '
                    + 'fo:clip="rect(%fin %fin %fin %fin)" ' % style[1]
                    + 'draw:luminance="0%" '
                    + 'draw:contrast="0" '
                    + 'draw:red="0%" '
                    + 'draw:green="0%" '
                    + 'draw:blue="0%" '
                    + 'draw:gamma="1" '
                    + 'draw:color-inversion="false" '
                    + 'draw:transparency="-100%" '
                    + 'draw:color-mode="standard"/>'
                    + "</style:style>\n"
                )
            elif style[0] == "Single":
                wrt2(
                    "<style:style "
                    + 'style:name="Single_%s" ' % str(style[1])
                    + 'style:family="graphic" '
                    + 'style:parent-style-name="photo">'
                    + "<style:graphic-properties "
                    + 'style:vertical-pos="from-top" '
                    + 'style:mirror="none" '
                    + 'fo:clip="rect(%fin %fin %fin %fin)" ' % style[1]
                    + 'draw:luminance="0%" '
                    + 'draw:contrast="0" '
                    + 'draw:red="0%" '
                    + 'draw:green="0%" '
                    + 'draw:blue="0%" '
                    + 'draw:gamma="1" '
                    + 'draw:color-inversion="false" '
                    + 'draw:transparency="-100%" '
                    + 'draw:color-mode="standard"/>'
                    + "</style:style>\n"
                )
            else:
                wrt2(
                    "<style:style "
                    + 'style:name="Row_%s" ' % str(style[1])
                    + 'style:family="graphic" '
                    + 'style:parent-style-name="Graphics">'
                    + "<style:graphic-properties "
                    + 'style:vertical-pos="from-top" '
                    + 'style:vertical-rel="paragraph" '
                    + 'style:horizontal-pos="from-left" '
                    + 'style:horizontal-rel="paragraph" '
                    + 'style:mirror="none" '
                    + 'fo:clip="rect(%fin %fin %fin %fin)" ' % style[1]
                    + 'draw:luminance="0%" '
                    + 'draw:contrast="0" '
                    + 'draw:red="0%" '
                    + 'draw:green="0%" '
                    + 'draw:blue="0%" '
                    + 'draw:gamma="1" '
                    + 'draw:color-inversion="false" '
                    + 'draw:transparency="-100%" '
                    + 'draw:color-mode="standard"/>'
                    + "</style:style>\n"
                )

    def add_media(self, file_name, pos, x_cm, y_cm, alt="", style_name=None, crop=None):
        """
        Add multi-media documents : photos
        """

        # try to open the image. If the open fails, it probably wasn't
        # a valid image (could be a PDF, or a non-image)
        (x, y) = image_size(file_name)
        if (x, y) == (0, 0):
            return

        not_extension, extension = os.path.splitext(file_name)
        file_name_hash = file_name
        file_name_hash = file_name_hash.encode("utf-8")
        odf_name = md5(file_name_hash).hexdigest() + extension

        media_list_item = (file_name, odf_name)
        if media_list_item not in self.media_list:
            self.media_list.append(media_list_item)

        base = escape(os.path.basename(file_name))
        tag = base.replace(".", "_")

        if self.new_cell:
            self.cntnt.write("<text:p>")

        pos = pos.title() if pos in ["left", "right", "single"] else "Row"

        if crop:
            (start_x, start_y, end_x, end_y) = crop_percentage_to_subpixel(x, y, crop)

            # Need to keep the ratio intact, otherwise scaled images look
            # stretched if the dimensions aren't close in size
            (act_width, act_height) = image_actual_size(
                x_cm, y_cm, int(end_x - start_x), int(end_y - start_y)
            )

            dpi = image_dpi(file_name)

            # ODF wants crop measurements in inch and as margins from each side
            left = start_x / dpi[0]
            right = (x - end_x) / dpi[0]
            top = start_y / dpi[1]
            bottom = (y - end_y) / dpi[1]
            crop = (top, right, bottom, left)

            self.stylelist_photos.append([pos, crop])

            pos += "_" + str(crop)

        else:
            (act_width, act_height) = image_actual_size(x_cm, y_cm, x, y)

        if len(alt):
            self.cntnt.write(
                '<draw:frame draw:style-name="%s" ' % pos
                + 'draw:name="caption_%s" ' % tag
                + 'text:anchor-type="paragraph" '
                + 'svg:y="0in" '
                + 'svg:width="%.2fcm" ' % act_width
                + 'draw:z-index="34"> '
                + '<draw:text-box fo:min-height="%.2fcm"> ' % act_height
                + '<text:p text:style-name="%s">' % style_name
            )

        self.cntnt.write(
            '<draw:frame draw:style-name="%s" ' % pos
            + 'draw:name="%s" ' % tag
            + 'text:anchor-type="paragraph" '
            + 'svg:width="%.2fcm" ' % act_width
            + 'svg:height="%.2fcm" ' % act_height
            + 'draw:z-index="1" >'
            + '<draw:image xlink:href="Pictures/%s" ' % odf_name
            + 'xlink:type="simple" xlink:show="embed" '
            + 'xlink:actuate="onLoad"/>\n'
            + "</draw:frame>\n"
        )

        if len(alt):
            self.cntnt.write(
                "%s" % "<text:line-break/>".join(alt)
                + "</text:p>"
                + "</draw:text-box>"
                + "</draw:frame>"
            )

        if self.new_cell:
            self.cntnt.write("</text:p>\n")

    def start_table(self, name, style_name):
        """
        open a table
        """
        self.cntnt.write(
            '<table:table table:name="%s" ' % name
            + 'table:style-name="%s">\n' % style_name
        )

        styles = self.get_style_sheet()
        table = styles.get_table_style(style_name)

        self.column_order = []
        for cell in range(table.get_columns()):
            self.column_order.append(cell)
        if self.get_rtl_doc():
            self.column_order.reverse()
        for col in self.column_order:
            self.cntnt.write(
                '<table:table-column table:style-name="%s.%s"/>\n'
                % (style_name, str(chr(ord("A") + col)))
            )

    def end_table(self):
        """
        close a table
        """
        self.cntnt.write("</table:table>\n")

    def start_row(self):
        """
        open a row
        """
        self.cntnt.write("<table:table-row>\n")
        self.row_cells = []
        self.cntnt_saved = self.cntnt  # save the content up to now

    def end_row(self):
        """
        close a row
        """
        self.cntnt = self.cntnt_saved  # restore the original contents
        if self.get_rtl_doc():
            self.row_cells.reverse()
        for cell in self.row_cells:
            self.cntnt.write(cell)
        self.cntnt.write("</table:table-row>\n")

    def start_cell(self, style_name, span=1):
        """
        open a cell
        """
        self.cntnt = StringIO()  # start a new buffer (with the expected name)
        self.span = span
        self.cntnt.write(
            '<table:table-cell table:style-name="%s" ' % style_name
            + 'table:value-type="string"'
        )
        if span > 1:
            self.cntnt.write(' table:number-columns-spanned="%s">\n' % span)
        else:
            self.cntnt.write(">\n")
        self.new_cell = 1

    def end_cell(self):
        """
        close a cell
        """
        self.cntnt.write("</table:table-cell>\n")
        # for col in range(1, self.span):
        #    self.cntnt.write('<table:covered-table-cell/>\n')
        self.new_cell = 0
        self.row_cells.append(self.cntnt.getvalue())  # save the cell, for now

    def start_bold(self):
        """
        open bold
        """
        self.cntnt.write('<text:span text:style-name="Tbold">')

    def end_bold(self):
        """
        close bold
        """
        self.cntnt.write("</text:span>")

    def start_superscript(self):
        """
        open superscript
        """
        self.cntnt.write('<text:span text:style-name="GSuper">')

    def end_superscript(self):
        """
        close superscript
        """
        self.cntnt.write("</text:span>")

    def _add_zip(self, zfile, name, data, date_time):
        """
        Add a zip file to an archive
        """
        zipinfo = zipfile.ZipInfo(name)
        zipinfo.date_time = date_time
        zipinfo.compress_type = zipfile.ZIP_DEFLATED
        zipinfo.external_attr = 0o644 << 16
        zfile.writestr(zipinfo, data)

    def _write_zip(self):
        """
        Create the odt file. This is a zip file
        """
        try:
            zfile = zipfile.ZipFile(self.filename, "w", zipfile.ZIP_DEFLATED)
        except IOError as msg:
            errmsg = "%s\n%s" % (_("Could not create %s") % self.filename, msg)
            raise ReportError(errmsg)
        except:
            raise ReportError(_("Could not create %s") % self.filename)

        now = time.localtime(time.time())[:6]

        self._add_zip(zfile, "META-INF/manifest.xml", self.mfile.getvalue(), now)
        self._add_zip(zfile, "content.xml", self.cntntx.getvalue(), now)
        self._add_zip(zfile, "meta.xml", self.meta.getvalue(), now)
        self._add_zip(zfile, "settings.xml", self.stfile.getvalue(), now)
        self._add_zip(zfile, "styles.xml", self.sfile.getvalue(), now)
        self._add_zip(zfile, "mimetype", self.mimetype.getvalue(), now)

        self.mfile.close()
        self.cntnt.close()
        self.meta.close()
        self.stfile.close()
        self.sfile.close()
        self.mimetype.close()

        for image in self.media_list:
            try:
                with open(image[0], mode="rb") as ifile:
                    self._add_zip(zfile, "Pictures/%s" % image[1], ifile.read(), now)
            except OSError as msg:
                errmsg = "%s\n%s" % (_("Could not open %s") % image[0], msg)
                raise ReportError(errmsg)
        zfile.close()

    def _write_styles_file(self):
        """
        create the styles.xml file
        """
        self.sfile = StringIO()
        wrtf = self.sfile.write

        wrtf('<?xml version="1.0" encoding="UTF-8"?>\n')
        wrtf("<office:document-styles " + _XMLNS + 'office:version="1.0">\n')

        wrtf("<office:font-face-decls>\n" + _FONTS + "</office:font-face-decls>\n")

        wrtf("<office:styles>\n" + _STYLES)

        styles = self.get_style_sheet()

        for style_name in styles.get_paragraph_style_names():
            style = styles.get_paragraph_style(style_name)
            wrtf(
                '<style:style style:name="%s" ' % style_name
                + 'style:family="paragraph" '
                + 'style:parent-style-name="Standard" '
                + 'style:class="text">\n'
                + "<style:paragraph-properties\n"
                + 'fo:margin-left="%.2fcm"\n' % style.get_left_margin()
                + 'fo:margin-right="%.2fcm"\n' % style.get_right_margin()
                + 'fo:margin-top="%.2fcm"\n' % style.get_top_margin()
                + 'fo:margin-bottom="%.2fcm"\n' % style.get_bottom_margin()
            )

            if style.get_padding() != 0.0:
                wrtf('fo:padding="%.2fcm" ' % style.get_padding())
            if style.get_header_level() > 0:
                wrtf('fo:keep-with-next="auto" ')

            align = style.get_alignment()
            if align == PARA_ALIGN_LEFT:
                wrtf('fo:text-align="start" ' 'style:justify-single-word="false" ')

            elif align == PARA_ALIGN_RIGHT:
                wrtf('fo:text-align="end" ')
            elif align == PARA_ALIGN_CENTER:
                wrtf('fo:text-align="center" ' 'style:justify-single-word="false" ')
            else:
                wrtf('fo:text-align="justify" ' 'style:justify-single-word="false" ')

            wrtf(
                'fo:text-indent="%.2fcm" ' % style.get_first_indent()
                + 'style:auto-text-indent="false"/> '
                + "<style:text-properties "
            )

            font = style.get_font()
            color = font.get_color()
            wrtf('fo:color="#%02x%02x%02x" ' % color)

            wrtf(
                'style:font-name="%s" '
                % (
                    "Arial"
                    if font.get_type_face() == FONT_SANS_SERIF
                    else "Times New Roman"
                )
            )

            wrtf('fo:font-size="%.0fpt" ' % font.get_size())
            if font.get_italic():
                wrtf('fo:font-style="italic" ')
            if font.get_bold():
                wrtf('fo:font-weight="bold" ')

            if font.get_underline():
                wrtf(
                    'style:text-underline="single" '
                    + 'style:text-underline-color="font-color" '
                    + 'fo:text-indent="%.2fcm" ' % style.get_first_indent()
                    + 'fo:margin-right="%.2fcm" ' % style.get_right_margin()
                    + 'fo:margin-left="%.2fcm" ' % style.get_left_margin()
                    + 'fo:margin-top="%.2fcm" ' % style.get_top_margin()
                    + 'fo:margin-bottom="%.2fcm"\n' % style.get_bottom_margin()
                )
            wrtf("/>\n" "</style:style>\n")

        # Dash lengths are based on the OpenOffice Ultrafine Dashed line style.
        for line_style in graphicstyle.line_style_names:
            dash_array = graphicstyle.get_line_style_by_name(line_style)
            wrtf(
                '<draw:stroke-dash draw:name="gramps_%s" draw:style="rect" '
                'draw:dots1="%d" draw:dots1-length="0.102cm" '
                'draw:dots2="%d" draw:dots2-length="0.102cm" '
                'draw:distance="%5.3fcm" />\n'
                % (line_style, dash_array[0], dash_array[0], dash_array[1] * 0.051)
            )

        # Current no leading number format for headers

        # wrtf('<text:outline-style>\n')
        # wrtf('<text:outline-level-style ')
        # wrtf('text:level="1" style:num-format=""/>\n')
        # wrtf('<text:outline-level-style ')
        # wrtf('text:level="2" style:num-format=""/>\n')
        # wrtf('<text:outline-level-style ')
        # wrtf('text:level="3" style:num-format=""/>\n')
        # wrtf('<text:outline-level-style ')
        # wrtf('text:level="4" style:num-format=""/>\n')
        # wrtf('<text:outline-level-style ')
        # wrtf('text:level="5" style:num-format=""/>\n')
        # wrtf('<text:outline-level-style ')
        # wrtf('text:level="6" style:num-format=""/>\n')
        # wrtf('<text:outline-level-style ')
        # wrtf('text:level="7" style:num-format=""/>\n')
        # wrtf('<text:outline-level-style ')
        # wrtf('text:level="8" style:num-format=""/>\n')
        # wrtf('<text:outline-level-style ')
        # wrtf('text:level="9" style:num-format=""/>\n')
        # wrtf('<text:outline-level-style ')
        # wrtf('text:level="10" style:num-format=""/>\n')
        # wrtf('</text:outline-style>\n')

        wrtf(
            "<text:notes-configuration  "
            'text:note-class="footnote"  '
            'style:num-format="1"  '
            'text:start-value="0"  '
            'text:footnotes-position="page"  '
            'text:start-numbering-at="document"/> '
        )

        wrtf(
            "<text:notes-configuration  "
            'text:note-class="endnote"  '
            'style:num-format="i"  '
            'text:start-value="0"/> '
        )

        wrtf(
            "<text:linenumbering-configuration  "
            'text:number-lines="false"  '
            'text:offset="0.499cm"  '
            'style:num-format="1"  '
            'text:number-position="left"  '
            'text:increment="5"/> '
        )

        wrtf("</office:styles>\n")
        wrtf(
            "<office:automatic-styles>\n"
            + _SHEADER_FOOTER
            + '<style:page-layout style:name="pm1">\n'
            + "<style:page-layout-properties "
            + 'fo:page-width="%.2fcm" ' % self.paper.get_size().get_width()
            + 'fo:page-height="%.2fcm" ' % self.paper.get_size().get_height()
            + 'style:num-format="1" '
        )

        wrtf(
            'style:print-orientation="%s" '
            % (
                "portrait"
                if self.paper.get_orientation() == PAPER_PORTRAIT
                else "landscape"
            )
        )

        wrtf(
            'fo:margin-top="%.2fcm" ' % self.paper.get_top_margin()
            + 'fo:margin-bottom="%.2fcm" ' % self.paper.get_bottom_margin()
            + 'fo:margin-left="%.2fcm" ' % self.paper.get_left_margin()
            + 'fo:margin-right="%.2fcm" ' % self.paper.get_right_margin()
            + 'style:writing-mode="lr-tb" '
            + 'style:footnote-max-height="0cm">\n'
            + '<style:footnote-sep style:width="0.018cm" '
            + 'style:distance-before-sep="0.101cm" '
            + 'style:distance-after-sep="0.101cm" '
            + 'style:adjustment="left" style:rel-width="25%" '
            + 'style:color="#000000"/>\n'
            + "</style:page-layout-properties>\n"
        )

        # header
        wrtf(
            "<style:header-style>\n"
            "<style:header-footer-properties "
            'fo:min-height="0cm" fo:margin-bottom="0.499cm"/>\n'
            "</style:header-style>\n"
        )

        # footer
        wrtf(
            "<style:footer-style>\n"
            "<style:header-footer-properties "
            'fo:min-height="0cm" fo:margin-bottom="0.499cm"/>\n'
            "</style:footer-style>\n"
        )

        # End of page layout
        wrtf("</style:page-layout>\n" "</office:automatic-styles>\n")

        # Master Styles
        wrtf(
            "<office:master-styles>\n"
            '<style:master-page style:name="Standard" '
            'style:page-layout-name="pm1">\n'
            # header
            #'<style:header>'
            #'<text:p text:style-name="S-Header">'
            # How to get the document title here ?
            #' TITRE : %s' % self.title
            #'</text:p>'
            #'</style:header>'
            # footer
            #'<style:footer>'
            #'<text:p text:style-name="S-Footer">'
            #'<text:page-number text:select-page="current">1'
            #'</text:page-number>/'
            #'<text:page-count>1'
            #'</text:page-count>'
            #'</text:p>'
            #'</style:footer>'
            #
            "</style:master-page>"
            "</office:master-styles>\n"
        )
        # End of document styles
        wrtf("</office:document-styles>\n")

    def page_break(self):
        """
        prepare a new page
        """
        self.new_page = 1

    def start_page(self):
        """
        create a new page
        """
        self.cntnt.write('<text:p text:style-name="docgen_page_break">\n')

    def end_page(self):
        """
        close the page
        """
        self.cntnt.write("</text:p>\n")

    def start_paragraph(self, style_name, leader=None):
        """
        open a new paragraph
        """
        style_sheet = self.get_style_sheet()
        style = style_sheet.get_paragraph_style(style_name)
        self.level = style.get_header_level()
        if self.new_page == 1:
            self.new_page = 0
            name = "NL%s" % style_name
        else:
            name = style_name
        if self.level == 0:
            self.cntnt.write('<text:p text:style-name="%s">' % name)
        else:
            self.cntnt.write(
                '<text:h text:style-name="%s"' % name
                + ' text:outline-level="%s">' % str(self.level)
            )
        if leader is not None:
            self.cntnt.write(leader + "<text:tab/>")
        self.new_cell = 0

    def end_paragraph(self):
        """
        close a paragraph
        """
        self.cntnt.write("</text:%s>\n" % ("p" if self.level == 0 else "h"))
        self.new_cell = 1

    def write_styled_note(
        self, styledtext, format, style_name, contains_html=False, links=False
    ):
        """
        Convenience function to write a styledtext to the ODF doc.
        styledtext : assumed a StyledText object to write
        format : = 0 : Flowed, = 1 : Preformatted
        style_name : name of the style to use for default presentation
        contains_html: bool, the backend should not check if html is present.
            If contains_html=True, then the textdoc is free to handle that in
            some way. Eg, a textdoc could remove all tags, or could make sure
            a link is clickable. ODFDoc prints the html without handling it
        links: bool, make URLs clickable if True
        """
        text = str(styledtext)
        s_tags = styledtext.get_tags()
        markuptext = self._backend.add_markup_from_styled(text, s_tags, "\n")

        if links is True:
            markuptext = re.sub(URL_PATTERN, _CLICKABLE, markuptext)

        # we need to know if we have new styles to add.
        # if markuptext contains : FontColor, FontFace, FontSize ...
        # we must prepare the new styles for the styles.xml file.
        # We are looking for the following format :
        # style-name="([a-zA-Z0-9]*)__([a-zA-Z0-9 ])">
        # The first element is the StyleType and the second one is the value
        start = 0
        while 1:
            m = NEW_STYLE.search(markuptext, start)
            if not m:
                break
            self.stylelist_notes.append(
                [m.group(1) + m.group(2), m.group(1), m.group(2)]
            )
            start = m.end()
        linenb = 1
        self.start_paragraph(style_name)
        for line in markuptext.split("\n"):
            [line, sigcount] = process_spaces(line, format)
            if sigcount == 0:
                self.end_paragraph()
                self.start_paragraph(style_name)
                linenb = 1
            else:
                if linenb > 1:
                    self.cntnt.write("<text:line-break/>")
                self.cntnt.write(line)
                linenb += 1
        self.end_paragraph()

    def write_text(self, text, mark=None, links=False, markup=False):
        """
        Uses the xml.sax.saxutils.escape function to convert XML
        entities. The ESC_MAP dictionary allows us to add our own
        mappings.
        @param mark:  IndexMark to use for indexing
        """
        if not markup:
            text = escape(text, ESC_MAP)

        if links is True:
            text = re.sub(URL_PATTERN, _CLICKABLE, text)

        self._write_mark(mark, text)

        self.cntnt.write(text)

    def write_markup(self, text, s_tags, mark=None):
        """
        Writes the text in the current paragraph.  Should only be used after a
        start_paragraph and before an end_paragraph.

        @param text: text to write. The text is assumed to be _not_ escaped
        @param s_tags:  assumed to be list of styledtexttags to apply to the
                        text
        @param mark:  IndexMark to use for indexing
        """
        markuptext = self._backend.add_markup_from_styled(text, s_tags)
        self.write_text(markuptext, mark=mark, markup=True)

    def _write_mark(self, mark, text):
        """
        Insert a mark at this point in the document.
        """
        if mark:
            key = escape(mark.key)
            key = key.replace('"', "&quot;")
            if mark.type == INDEX_TYPE_ALP:
                self.cntnt.write(
                    "<text:alphabetical-index-mark " 'text:string-value="%s" />' % key
                )
            elif mark.type == INDEX_TYPE_TOC:
                self.cntnt.write(
                    "<text:toc-mark "
                    + 'text:string-value="%s" ' % key
                    + 'text:outline-level="%d" />' % mark.level
                )
            elif mark.type == LOCAL_HYPERLINK:
                self.cntnt.write('<text:a xlink:type="simple" xlink:href="%s">' % key)
                self.cntnt.write(text)
                self.cntnt.write("</text:a>")
                return
            elif mark.type == LOCAL_TARGET:
                self.cntnt.write('<text:bookmark text:name="%s"/>' % key)

    def insert_toc(self):
        """
        Insert a Table of Contents at this point in the document.
        """
        title = self.toc_title
        self.cntnt.write("<text:table-of-content>")

        self.cntnt.write(
            "<text:table-of-content-source "
            + 'text:outline-level="3" '
            + 'text:use-outline-level="false">'
        )

        self.cntnt.write(
            "<text:index-title-template " + 'text:style-name="TOC-Title">' + title
        )
        self.cntnt.write("</text:index-title-template>")

        for level in range(1, 4):
            self.cntnt.write(
                "<text:table-of-content-entry-template "
                + 'text:outline-level="%d" ' % level
                + 'text:style-name="TOC-Heading%d">' % level
            )

            self.cntnt.write("<text:index-entry-chapter/>")
            self.cntnt.write("<text:index-entry-text/>")
            self.cntnt.write(
                "<text:index-entry-tab-stop "
                + 'style:type="right" '
                + 'style:leader-char="."/>'
            )
            self.cntnt.write("<text:index-entry-page-number/>")
            self.cntnt.write("</text:table-of-content-entry-template>")

        self.cntnt.write("</text:table-of-content-source>")

        self.cntnt.write("<text:index-body>")
        self.cntnt.write("<text:index-title>")
        self.cntnt.write('<text:p text:style-name="NLTOC-Title">%s</text:p>' % title)
        self.cntnt.write("</text:index-title>")
        self.cntnt.write("</text:index-body>")

        self.cntnt.write("</text:table-of-content>")

    def insert_index(self):
        """
        Insert an Alphabetical Index at this point in the document.
        """
        title = self.index_title
        self.cntnt.write("<text:alphabetical-index>")
        self.cntnt.write(
            "<text:alphabetical-index-source "
            + 'text:ignore-case="true"  '
            + 'text:combine-entries="false"  '
            + 'text:combineentries-with-pp="false">'
        )

        self.cntnt.write(
            "<text:index-title-template " + 'text:style-name="IDX-Title">' + title
        )
        self.cntnt.write("</text:index-title-template>")

        self.cntnt.write(
            "<text:alphabetical-index-entry-template "
            + 'text:outline-level="1" '
            + 'text:style-name="IDX-Entry">'
        )
        self.cntnt.write("<text:index-entry-text/>")
        self.cntnt.write(
            "<text:index-entry-tab-stop "
            + 'style:type="right" '
            + 'style:leader-char="."/>'
        )
        self.cntnt.write("<text:index-entry-page-number/>")
        self.cntnt.write("</text:alphabetical-index-entry-template>")

        self.cntnt.write("</text:alphabetical-index-source>")

        self.cntnt.write("<text:index-body>")
        self.cntnt.write("<text:index-title>")
        self.cntnt.write('<text:p text:style-name="NLIDX-Title">%s</text:p>' % title)
        self.cntnt.write("</text:index-title>")
        self.cntnt.write("</text:index-body>")

        self.cntnt.write("</text:alphabetical-index>")

    def _write_manifest(self):
        """
        create the manifest.xml file
        """
        self.mfile = StringIO()

        # Header
        self.mfile.write(
            '<?xml version="1.0" encoding="UTF-8"?>\n'
            + "<manifest:manifest "
            + 'xmlns:manifest="urn:oasis:names:tc:opendocument'
            + ':xmlns:manifest:1.0">'
            + "<manifest:file-entry "
            + 'manifest:media-type="%s" ' % APP_TYPE
            + 'manifest:full-path="/"/>'
        )

        # Images
        for image in self.media_list:
            self.mfile.write(
                '<manifest:file-entry manifest:media-type="" '
                + 'manifest:full-path="Pictures/'
                + image[1]
                + '"/>'
            )

        # Footer
        self.mfile.write(
            '<manifest:file-entry manifest:media-type="" '
            'manifest:full-path="Pictures/"/>'
            '<manifest:file-entry manifest:media-type="text/xml" '
            'manifest:full-path="content.xml"/>'
            '<manifest:file-entry manifest:media-type="text/xml" '
            'manifest:full-path="styles.xml"/>'
            '<manifest:file-entry manifest:media-type="text/xml" '
            'manifest:full-path="settings.xml"/>'
            '<manifest:file-entry manifest:media-type="text/xml" '
            'manifest:full-path="meta.xml"/>'
            "</manifest:manifest>\n"
        )

    def _write_settings(self):
        """
        create the settings.xml file
        """
        self.stfile = StringIO()
        # This minimal settings file has been taken from
        # http://mashupguide.net/1.0/html/ch17s03.xhtml (Creative commons
        # licence): http://mashupguide.net/1.0/html/apas02.xhtml
        self.stfile.write(
            '<?xml version="1.0" encoding="UTF-8"?>\n'
            + '<office:document-settings office:version="1.0"\n'
            + 'xmlns:config="urn:oasis:names:tc:opendocument:xmlns:config:1.0"\n'
            + 'xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"\n'
            + 'xmlns:ooo="http://openoffice.org/2004/office"\n'
            + 'xmlns:xlink="http://www.w3.org/1999/xlink" />'
        )

    def _write_mimetype_file(self):
        """
        create the mimetype.xml file
        """
        self.mimetype = StringIO()
        self.mimetype.write("application/vnd.oasis.opendocument.text")

    def _write_meta_file(self):
        """
        create the meta.xml file
        """
        self.meta = StringIO()
        generator = PROGRAM_NAME + " " + VERSION
        creator = escape(self.get_creator(), ESC_MAP)
        date = self.time
        lang = self.lang
        gramps_home_url = URL_HOMEPAGE

        self.meta.write(_META_XML % locals())

    def rotate_text(self, style, text, x, y, angle, mark=None):
        """
        Used to rotate a text with an angle.
        @param mark:  IndexMark to use for indexing
        """
        style_sheet = self.get_style_sheet()
        stype = style_sheet.get_draw_style(style)
        pname = stype.get_paragraph_style()
        para = style_sheet.get_paragraph_style(pname)
        font = para.get_font()
        size = font.get_size()

        height = size * (len(text))
        width = 0
        for line in text:
            width = max(width, string_width(font, line))
        wcm = utils.pt2cm(width)
        hcm = utils.pt2cm(height)

        rangle = radians(angle)
        xloc = x - (wcm / 2.0) * cos(rangle) + (hcm / 2.0) * sin(rangle)
        yloc = y - (hcm / 2.0) * cos(rangle) - (wcm / 2.0) * sin(rangle)

        self._write_mark(mark, text)

        self.cntnt.write(
            '<draw:frame text:anchor-type="paragraph" '
            + 'draw:z-index="2" '
            + 'draw:style-name="clear" '
            + 'svg:height="%.2fcm" ' % hcm
            + 'svg:width="%.2fcm" ' % wcm
            + 'draw:transform="'
            + "rotate (%.8f) " % -rangle
            + 'translate (%.3fcm %.3fcm)">\n' % (xloc, yloc)
            + "<draw:text-box>\n"
            + '<text:p text:style-name="X%s">' % pname
            + '<text:span text:style-name="F%s">' % pname
            + escape("\n".join(text), ESC_MAP)
            + "</text:span></text:p>\n</draw:text-box>\n"
            + "</draw:frame>\n"
        )

    def draw_path(self, style, path):
        """
        Draw a path
        """
        minx = 9e12
        miny = 9e12
        maxx = 0
        maxy = 0

        for point in path:
            minx = min(point[0], minx)
            miny = min(point[1], miny)
            maxx = max(point[0], maxx)
            maxy = max(point[1], maxy)

        self.cntnt.write(
            '<draw:polygon draw:style-name="%s" ' % style
            + 'draw:layer="layout" '
            + 'draw:z-index="1" '
            + 'svg:x="%2fcm" svg:y="%2fcm" ' % (float(minx), float(miny))
            + 'svg:viewBox="0 0 %d %d" '
            % (int((maxx - minx) * 1000), int((maxy - miny) * 1000))
            + 'svg:width="%.4fcm" ' % (maxx - minx)
            + 'svg:height="%.4fcm" ' % (maxy - miny)
        )

        point = path[0]
        x1 = int((point[0] - minx) * 1000)
        y1 = int((point[1] - miny) * 1000)
        self.cntnt.write('draw:points="%d, %d' % (x1, y1))

        for point in path[1:]:
            x1 = int((point[0] - minx) * 1000)
            y1 = int((point[1] - miny) * 1000)
            self.cntnt.write(" %d, %d" % (x1, y1))
        self.cntnt.write('"/>\n')

    def draw_line(self, style, x1, y1, x2, y2):
        """
        Draw a line
        """
        self.cntnt.write(
            '<draw:line text:anchor-type="paragraph" '
            + 'draw:z-index="3" '
            + 'draw:style-name="%s" ' % style
            + 'svg:x1="%.2fcm" ' % x1
            + 'svg:y1="%.2fcm" ' % y1
            + 'svg:x2="%.2fcm" ' % x2
            + 'svg:y2="%.2fcm">' % y2
            + "<text:p/>\n"
            + "</draw:line>\n"
        )

    def draw_text(self, style, text, x, y, mark=None):
        """
        Draw a text
        @param mark:  IndexMark to use for indexing
        """
        style_sheet = self.get_style_sheet()
        box_style = style_sheet.get_draw_style(style)
        para_name = box_style.get_paragraph_style()
        pstyle = style_sheet.get_paragraph_style(para_name)
        font = pstyle.get_font()
        sw = utils.pt2cm(string_width(font, text)) * 1.3

        self._write_mark(mark, text)

        self.cntnt.write(
            '<draw:frame text:anchor-type="paragraph" '
            + 'draw:z-index="2" '
            + 'draw:style-name="%s" ' % style
            + 'svg:width="%.2fcm" ' % sw
            + 'svg:height="%.2fcm" ' % (utils.pt2cm(font.get_size() * 1.4))
            + 'svg:x="%.2fcm" ' % float(x)
            + 'svg:y="%.2fcm">' % float(y)
            + "<draw:text-box> "
            + '<text:p text:style-name="F%s">' % para_name
            + '<text:span text:style-name="F%s">' % para_name
            +
            #' fo:max-height="%.2f">' % font.get_size()  +
            escape(text, ESC_MAP)
            + "</text:span>"
            + "</text:p>"
            + "</draw:text-box>\n"
            + "</draw:frame>\n"
        )

    def draw_box(self, style, text, x, y, w, h, mark=None):
        """
        Draw a box
        @param mark:  IndexMark to use for indexing
        """
        style_sheet = self.get_style_sheet()
        box_style = style_sheet.get_draw_style(style)
        para_name = box_style.get_paragraph_style()
        shadow_width = box_style.get_shadow_space()

        self._write_mark(mark, text)

        if box_style.get_shadow():
            self.cntnt.write(
                '<draw:rect text:anchor-type="paragraph" '
                + 'draw:style-name="%s_shadow" ' % style
                + 'draw:z-index="0" '
                + 'draw:text-style-name="%s" ' % para_name
                + 'svg:width="%.2fcm" ' % w
                + 'svg:height="%.2fcm" ' % h
                + 'svg:x="%.2fcm" ' % (float(x) + shadow_width)
                + 'svg:y="%.2fcm">\n' % (float(y) + shadow_width)
                + "</draw:rect>\n"
            )

        self.cntnt.write(
            '<draw:rect text:anchor-type="paragraph" '
            + 'draw:style-name="%s" ' % style
            + 'draw:text-style-name="%s" ' % para_name
            + 'draw:z-index="1" '
            + 'svg:width="%.2fcm" ' % w
            + 'svg:height="%.2fcm" ' % h
            + 'svg:x="%.2fcm" ' % float(x)
            + 'svg:y="%.2fcm">\n' % float(y)
        )
        if text:
            self.cntnt.write(
                '<text:p text:style-name="%s">' % para_name
                + '<text:span text:style-name="F%s">' % para_name
                + escape(text, ESC_MAP)
                + "</text:span>"
                "</text:p>\n"
            )
        self.cntnt.write("</draw:rect>\n")

    def center_text(self, style, text, x, y, mark=None):
        """
        Center a text in a cell, a row, a line, ...
        @param mark:  IndexMark to use for indexing
        """
        style_sheet = self.get_style_sheet()
        box_style = style_sheet.get_draw_style(style)
        para_name = box_style.get_paragraph_style()
        pstyle = style_sheet.get_paragraph_style(para_name)
        font = pstyle.get_font()

        size = (string_width(font, text) / 72.0) * 2.54

        self._write_mark(mark, text)

        self.cntnt.write(
            '<draw:frame text:anchor-type="paragraph" '
            + 'draw:style-name="%s" ' % style
            + 'draw:z-index="2" '
            + 'svg:width="%.2fcm" ' % size
            + 'svg:height="%.2fpt" ' % font.get_size()
            + 'svg:x="%.2fcm" ' % (x - (size / 2.0))
            + 'svg:y="%.2fcm">\n' % float(y)
        )

        if text:
            self.cntnt.write(
                "<draw:text-box>"
                + '<text:p text:style-name="X%s">' % para_name
                + '<text:span text:style-name="F%s">' % para_name
                + escape(text, ESC_MAP)
                + "</text:span>\n"
                + "</text:p>\n"
                + "</draw:text-box>"
            )
        self.cntnt.write("</draw:frame>\n")


def process_spaces(line, format):
    """
    Function to process spaces in text lines for flowed and pre-formatted notes.
    line : text to process
    format : = 0 : Flowed, = 1 : Preformatted

    If the text is flowed (format==0), then leading spaces (after ignoring XML)
    are removed. Embedded multiple spaces are reduced to one by ODF
    If the text is pre-formatted (format==1), then all spaces (after ignoring
    XML) are replaced by "<text:s/>"

    Returns the processed text, and the number of significant
    (i.e. non-white-space) chars.
    """
    txt = ""
    xml = False
    sigcount = 0
    # We loop through every character, which is very inefficient, but an attempt
    # to use a regex replace didn't always work. This was the code that was
    # replaced.
    # Problem, we may not replace ' ' in xml tags, so we use a regex
    # self.cntnt.write(re.sub(' (?=([^(<|>)]*<[^>]*>)*[^>]*$)',
    #                        "<text:s/>", line))
    for char in line:
        if char == "<" and xml is False:
            xml = True
            txt += char
        elif char == ">" and xml is True:
            xml = False
            txt += char
        elif xml is True:
            txt += char
        elif char == " " or char == "\t":
            if format == 0 and sigcount == 0:
                pass
            elif format == 1:
                # preformatted, section White-space characters of
                # http://docs.oasis-open.org/office/v1.1/OS/OpenDocument-v1.1-html/OpenDocument-v1.1.html#5.1.1.White-space%20Characters|outline
                txt += "<text:s/>"
            else:
                txt += char
        else:
            sigcount += 1
            txt += char
    return [txt, sigcount]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#               2009       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from .editsecondary import EditSecondary
from ..widgets import MonitoredEntry, PrivacyButton, MonitoredDataType
from ..glade import Glade
from gramps.gen.const import URL_MANUAL_SECT1

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT1
WIKI_HELP_SEC = _("Internet_Address_Editor", "manual")


# -------------------------------------------------------------------------
#
# EditUrl class
#
# -------------------------------------------------------------------------
class EditUrl(EditSecondary):
    def __init__(self, dbstate, uistate, track, name, url, callback):
        EditSecondary.__init__(self, dbstate, uistate, track, url, callback)

    def _local_init(self):
        self.top = Glade()
        self.jump = self.top.get_object("jump")

        self.set_window(
            self.top.toplevel,
            self.top.get_object("title"),
            _("Internet Address Editor"),
        )
        self.setup_configs("interface.url", 600, 150)

    def _connect_signals(self):
        self.jump.connect("clicked", self.jump_to)
        self.define_cancel_button(self.top.get_object("button125"))
        self.define_ok_button(self.top.get_object("button124"), self.save)
        # TODO help button (rename glade button name)
        self.define_help_button(
            self.top.get_object("button130"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

    def jump_to(self, obj):
        if self.obj.get_path():
            from ..display import display_url

            display_url(self.obj.get_path())

    def _setup_fields(self):
        self.des = MonitoredEntry(
            self.top.get_object("url_des"),
            self.obj.set_description,
            self.obj.get_description,
            self.db.readonly,
        )

        self.addr = MonitoredEntry(
            self.top.get_object("url_addr"),
            self.obj.set_path,
            self.obj.get_path,
            self.db.readonly,
        )

        self.priv = PrivacyButton(
            self.top.get_object("priv"), self.obj, self.db.readonly
        )

        self.type_sel = MonitoredDataType(
            self.top.get_object("type"),
            self.obj.set_type,
            self.obj.get_type,
            self.db.readonly,
            self.db.get_url_types(),
        )

    def build_menu_names(self, obj):
        etitle = _("Internet Address Editor")
        return (etitle, etitle)

    def save(self, *obj):
        self.callback(self.obj)
        self.close()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...datehandler import parser
from ...lib.eventtype import EventType
from . import Rule
from ...utils.db import get_participant_from_event
from ...display.place import displayer as place_displayer


# -------------------------------------------------------------------------
#
# HasEventBase
#
# -------------------------------------------------------------------------
class HasEventBase(Rule):
    """Rule that checks for an event with a particular value."""

    labels = ["Event type:", "Date:", "Place:", "Description:", "Main Participants:"]
    name = "Events matching parameters"
    description = "Matches events with particular parameters"
    category = _("Event filters")
    allow_regex = True

    def prepare(self, db, user):
        self.date = None
        if self.list[0]:
            self.etype = EventType()
            self.etype.set_from_xml_str(self.list[0])
        else:
            self.etype = None
        try:
            if self.list[1]:
                self.date = parser.parse(self.list[1])
        except:
            pass

    def apply(self, db, event):
        if self.etype:
            if self.etype.is_custom() and self.use_regex:
                if self.regex[0].search(str(event.type)) is None:
                    return False
            elif event.type != self.etype:
                return False

        if not self.match_substring(3, event.get_description()):
            return False

        if self.date:
            if not event.get_date_object().match(self.date):
                return False

        if self.list[2]:
            place_id = event.get_place_handle()
            if place_id:
                place = db.get_place_from_handle(place_id)
                place_title = place_displayer.display(db, place)
                if not self.match_substring(2, place_title):
                    return False
            else:
                return False

        if not self.match_substring(
            4, get_participant_from_event(db, event.get_handle(), all_=True)
        ):
            return False

        return True



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011 Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gui.listmodel import ListModel
from gramps.gen.utils.db import navigation_label
from gramps.gen.plug import Gramplet
from gramps.gui.utils import edit_object
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.datehandler import displayer

_ = glocale.translation.gettext


class Backlinks(Gramplet):
    """
    Displays the back references for an object.
    """

    def init(self):
        self.date_column = None
        self.evts = False
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)
        self.gui.WIDGET.show()

    def build_gui(self):
        """
        Build the GUI interface.
        """
        self.top = Gtk.TreeView()
        titles = [
            (_("Type"), 1, 100),
            (_("Name"), 2, 100),
            (_("Date"), 4, 200),
            ("sd", 4, 120),  # sorted date column
            ("", 5, 1),  # hidden column for the handle
            ("", 6, 1),  # hidden column for non-localized object type
        ]
        self.model = ListModel(self.top, titles, event_func=self.cb_double_click)
        self.date_column = self.top.get_column(2)
        self.sdate = self.top.get_column(3)
        self.top.get_column(1).set_expand(True)  # The name use the max
        # possible size
        return self.top

    def display_backlinks(self, active_handle):
        """
        Display the back references for an object.
        """
        self.evts = False
        sdcolumn = None
        for classname, handle in self.dbstate.db.find_backlink_handles(active_handle):
            name = navigation_label(self.dbstate.db, classname, handle)[0]
            sdcolumn = self.top.get_column(3)
            dcolumn = self.top.get_column(2)
            if classname == "Event":
                obj = self.dbstate.db.get_event_from_handle(handle)
                o_date = obj.get_date_object()
                date = displayer.display(o_date)
                sdate = "%09d" % o_date.get_sort_value()
                sdcolumn.set_sort_column_id(3)
                dcolumn.set_sort_column_id(3)
                self.evts = True
            else:
                sdcolumn.set_sort_column_id(1)
                date = sdate = ""
            self.model.add((_(classname), name, date, sdate, handle, classname))
        if self.evts:
            self.date_column.set_visible(True)
            sdcolumn.set_visible(False)
        else:
            self.date_column.set_visible(False)
            if sdcolumn:
                sdcolumn.set_visible(False)
        self.set_has_data(self.model.count > 0)

    def get_has_data(self, active_handle):
        """
        Return True if the gramplet has data, else return False.
        """
        if not active_handle:
            return False
        for handle in self.dbstate.db.find_backlink_handles(active_handle):
            return True
        return False

    def cb_double_click(self, treeview):
        """
        Handle double click on treeview.
        """
        (model, iter_) = treeview.get_selection().get_selected()
        if not iter_:
            return

        (objclass, handle) = (model.get_value(iter_, 5), model.get_value(iter_, 4))

        edit_object(self.dbstate, self.uistate, objclass, handle)

    def db_changed(self):
        for item in [
            "person",
            "family",
            "source",
            "citation",
            "event",
            "media",
            "place",
            "repository",
            "note",
        ]:
            self.connect(self.dbstate.db, "%s-delete" % item, self.update)
            self.connect(self.dbstate.db, "%s-add" % item, self.update)
            self.connect(self.dbstate.db, "%s-update" % item, self.update)


class PersonBacklinks(Backlinks):
    """
    Displays the back references for a person.
    """

    def active_changed(self, handle):
        self.update()

    def update_has_data(self):
        active_handle = self.get_active("Person")
        self.set_has_data(self.get_has_data(active_handle))

    def main(self):
        active_handle = self.get_active("Person")
        self.model.clear()
        if active_handle:
            self.display_backlinks(active_handle)
        else:
            self.set_has_data(False)


class EventBacklinks(Backlinks):
    """
    Displays the back references for an event.
    """

    def db_changed(self):
        super().db_changed()
        self.connect_signal("Event", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Event")
        self.set_has_data(self.get_has_data(active_handle))

    def main(self):
        active_handle = self.get_active("Event")
        self.model.clear()
        if active_handle:
            self.display_backlinks(active_handle)
        else:
            self.set_has_data(False)


class FamilyBacklinks(Backlinks):
    """
    Displays the back references for a family.
    """

    def db_changed(self):
        super().db_changed()
        self.connect_signal("Family", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Family")
        self.set_has_data(self.get_has_data(active_handle))

    def main(self):
        active_handle = self.get_active("Family")
        self.model.clear()
        if active_handle:
            self.display_backlinks(active_handle)
        else:
            self.set_has_data(False)


class PlaceBacklinks(Backlinks):
    """
    Displays the back references for a place.
    """

    def db_changed(self):
        super().db_changed()
        self.connect_signal("Place", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Place")
        self.set_has_data(self.get_has_data(active_handle))

    def main(self):
        active_handle = self.get_active("Place")
        self.model.clear()
        if active_handle:
            self.display_backlinks(active_handle)
        else:
            self.set_has_data(False)


class SourceBacklinks(Backlinks):
    """
    Displays the back references for a source,.
    """

    def db_changed(self):
        super().db_changed()
        self.connect_signal("Source", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Source")
        self.set_has_data(self.get_has_data(active_handle))

    def main(self):
        active_handle = self.get_active("Source")
        self.model.clear()
        if active_handle:
            self.display_backlinks(active_handle)
        else:
            self.set_has_data(False)


class CitationBacklinks(Backlinks):
    """
    Displays the back references for a Citation,.
    """

    def db_changed(self):
        super().db_changed()
        self.connect_signal("Citation", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Citation")
        self.set_has_data(self.get_has_data(active_handle))

    def main(self):
        active_handle = self.get_active("Citation")
        self.model.clear()
        if active_handle:
            self.display_backlinks(active_handle)
        else:
            self.set_has_data(False)


class RepositoryBacklinks(Backlinks):
    """
    Displays the back references for a repository.
    """

    def db_changed(self):
        super().db_changed()
        self.connect_signal("Repository", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Repository")
        self.set_has_data(self.get_has_data(active_handle))

    def main(self):
        active_handle = self.get_active("Repository")
        self.model.clear()
        if active_handle:
            self.display_backlinks(active_handle)
        else:
            self.set_has_data(False)


class MediaBacklinks(Backlinks):
    """
    Displays the back references for a media object.
    """

    def db_changed(self):
        super().db_changed()
        self.connect_signal("Media", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Media")
        self.set_has_data(self.get_has_data(active_handle))

    def main(self):
        active_handle = self.get_active("Media")
        self.model.clear()
        if active_handle:
            self.display_backlinks(active_handle)
        else:
            self.set_has_data(False)


class NoteBacklinks(Backlinks):
    """
    Displays the back references for a note.
    """

    def db_changed(self):
        super().db_changed()
        self.connect_signal("Note", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Note")
        self.set_has_data(self.get_has_data(active_handle))

    def main(self):
        active_handle = self.get_active("Note")
        self.model.clear()
        if active_handle:
            self.display_backlinks(active_handle)
        else:
            self.set_has_data(False)



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
# Copyright (C) 2018       Theo van Rijn
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    ThumbnailPreviewPage
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from decimal import getcontext
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.lib import Media
from gramps.plugins.lib.libhtml import Html

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.plugins.webreport.common import FULLCLEAR, html_escape

_ = glocale.translation.sgettext
LOG = logging.getLogger(".NarrativeWeb")
getcontext().prec = 8


class ThumbnailPreviewPage(BasePage):
    """
    This class is responsible for displaying information about
    the Thumbnails page.
    """

    def __init__(self, report, the_lang, the_title, cb_progress):
        """
        @param: report      -- The instance of the main report class
                               for this report
        @param: the_lang    -- Is the lang to process.
        @param: title       -- Is the title of the web page
        @param: cb_progress -- The step used for the progress bar.
        """
        BasePage.__init__(self, report, the_lang, the_title)
        self.create_thumbs_only = report.options["create_thumbs_only"]
        self.create_thumbs_index = self.report.options["create_thumbs_index"]

        # bug 8950 : it seems it's better to sort on desc + gid.
        def sort_by_desc_and_gid(obj):
            """
            Sort by media description and gramps ID
            """
            return (obj.desc, obj.gramps_id)

        self.photo_keys = sorted(
            self.report.obj_dict[Media],
            key=lambda x: sort_by_desc_and_gid(self.r_db.get_media_from_handle(x)),
        )

        if self.create_unused_media:
            # add unused media
            media_list = self.r_db.get_media_handles()
            for media_ref in media_list:
                if media_ref not in self.report.obj_dict[Media]:
                    self.photo_keys.append(media_ref)

        media_list = []
        for person_handle in self.photo_keys:
            photo = self.r_db.get_media_from_handle(person_handle)
            if photo:
                if photo.get_mime_type().startswith("image"):
                    media_list.append((photo.get_description(), person_handle, photo))

                    if self.create_thumbs_only:
                        self.copy_thumbnail(person_handle, photo)

        media_list.sort(key=lambda x: self.rlocale.sort_key(x[0]))

        # Create thumbnail preview page...
        if self.create_thumbs_index:
            output_file, sio = self.report.create_file("thumbnails")
        result = self.write_header(self._("Thumbnails"))
        thumbnailpage, dummy_head, body, outerwrapper = result

        with Html("div", class_="content", id="Preview") as previewpage:
            outerwrapper += previewpage

            msg = self._(
                "This page displays a indexed list "
                "of all the media objects "
                "in this database. It is sorted by media title. "
                "There is an index "
                "of all the media objects in this database. "
                "Clicking on a thumbnail "
                "will take you to that image&#8217;s page."
            )
            previewpage += Html("p", msg, id="description")

        with Html("div", id="gallery") as gallery:
            previewpage += gallery
            index, indexpos = 1, 0
            num_of_images = len(media_list)
            while index <= num_of_images:
                ptitle = media_list[indexpos][0]
                person_handle = media_list[indexpos][1]
                photo = media_list[indexpos][2]

                # begin table cell and attach to table row(trow)...
                gallerycell = Html("div", class_="gallerycell " + self.dir)
                gallery += gallerycell

                # attach index number...
                numberdiv = Html("div", class_="indexno")
                gallerycell += numberdiv

                # attach anchor name to date cell in upper right
                # corner of grid...
                numberdiv += Html("a", index, name=index, title=index, inline=True)

                # create thumbnail
                (dummy_real_path, newpath) = self.report.prepare_copy_media(photo)
                newpath = self.report.build_url_fname(newpath, image=True)
                newpathc = newpath

                # attach thumbnail to list...
                gallerycell += self.thumb_hyper_image(
                    newpathc, "img", person_handle, ptitle
                )

                index += 1
                indexpos += 1

        # begin Thumbnail Reference section...
        with Html("div", class_="content", id="references") as section:
            outerwrapper += section
            section += Html("h4", self._("References"), inline=True)

            with Html("table", class_="infolist") as table:
                section += table

                tbody = Html("tbody")
                table += tbody

                index = 1
                for ptitle, person_handle, photo in media_list:
                    trow = Html("tr")
                    tbody += trow

                    tcell1 = Html(
                        "td",
                        self.thumbnail_link(ptitle, index),
                        class_="ColumnRowLabel",
                    )
                    tcell2 = Html("td", ptitle, class_="ColumnName")
                    trow += (tcell1, tcell2)

                    # increase progress meter...
                    cb_progress()

                    # increase index for row number...
                    index += 1

        # add body id element
        body.attr = 'id ="ThumbnailPreview"'

        # add footer section
        # add clearline for proper styling
        footer = self.write_footer(None)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        if self.create_thumbs_index:
            self.xhtml_writer(thumbnailpage, output_file, sio, 0)

    def thumbnail_link(self, name, index):
        """
        creates a hyperlink for Thumbnail Preview Reference...

        @param: name    -- The image description
        @param: index   -- The image index
        """
        return Html("a", index, title=html_escape(name), href="#%d" % index)

    def thumb_hyper_image(self, thumbnail_url, subdir, fname, name):
        """
        replaces media_link() because it doesn't work for this instance

        @param: thumnail_url -- The url for this thumbnail
        @param: subdir       -- The subdir prefix to add
        @param: fname        -- The file name for this image
        @param: name         -- The image description
        """
        name = html_escape(name)
        url = "/".join(self.report.build_subdirs(subdir, fname) + [fname]) + self.ext
        with Html("div", class_="thumbnail") as thumbnail:
            # snapshot += thumbnail

            if not self.create_thumbs_only:
                thumbnail_link = Html("a", href=url, title=name) + (
                    Html("img", src=thumbnail_url, alt=name)
                )
            else:
                thumbnail_link = Html("img", src=thumbnail_url, alt=name)
            thumbnail += thumbnail_link
        return thumbnail



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Package providing filtering framework for Gramps.
"""


class SearchFilter:
    def __init__(self, func, text, invert):
        self.func = func
        self.text = text.upper()
        self.invert = invert

    def match(self, handle, db):
        return self.invert ^ (self.func(handle).upper().find(self.text) != -1)


class ExactSearchFilter(SearchFilter):
    def __init__(self, func, text, invert):
        SearchFilter.__init__(self, func, text, invert)

    def match(self, handle, db):
        return self.invert ^ (self.func(handle).upper() == self.text.strip())



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Tools/Database Processing/Find Possible Duplicate People"""

# -------------------------------------------------------------------------
#
# GNOME libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import URL_MANUAL_PAGE
from gramps.gen.lib import Event, Person
from gramps.gui.utils import ProgressMeter
from gramps.gui.plug import tool
from gramps.gen.soundex import soundex, compare
from gramps.gen.display.name import displayer as name_displayer
from gramps.gui.dialog import OkDialog
from gramps.gui.listmodel import ListModel
from gramps.gen.errors import WindowActiveError
from gramps.gui.merge import MergePerson
from gramps.gui.display import display_help
from gramps.gui.managedwindow import ManagedWindow
from gramps.gui.dialog import RunDatabaseRepair
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gui.glade import Glade

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
_val2label = {
    0.25: _("Low"),
    1.0: _("Medium"),
    2.0: _("High"),
}

WIKI_HELP_PAGE = "%s_-_Tools" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Find_Possible_Duplicate_People", "manual")


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
def is_initial(name):
    if len(name) > 2:
        return 0
    elif len(name) == 2:
        if name[0] == name[0].upper() and name[1] == ".":
            return 1
    else:
        return name[0] == name[0].upper()


# -------------------------------------------------------------------------
#
# The Actual tool.
#
# -------------------------------------------------------------------------
class DuplicatePeopleTool(tool.Tool, ManagedWindow):
    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate

        tool.Tool.__init__(self, dbstate, options_class, name)
        ManagedWindow.__init__(self, uistate, [], self.__class__)
        self.dbstate = dbstate
        self.uistate = uistate
        self.map = {}
        self.list = []
        self.index = 0
        self.merger = None
        self.mergee = None
        self.removed = {}
        self.update = callback
        self.use_soundex = 1

        top = Glade(toplevel="finddupes", also_load=["liststore1"])

        # retrieve options
        threshold = self.options.handler.options_dict["threshold"]
        use_soundex = self.options.handler.options_dict["soundex"]

        my_menu = Gtk.ListStore(str, object)
        for val in sorted(_val2label):
            my_menu.append([_val2label[val], val])

        self.soundex_obj = top.get_object("soundex")
        self.soundex_obj.set_active(use_soundex)
        self.soundex_obj.show()

        self.menu = top.get_object("menu")
        self.menu.set_model(my_menu)
        self.menu.set_active(0)

        window = top.toplevel
        self.set_window(
            window, top.get_object("title"), _("Find Possible Duplicate People")
        )
        self.setup_configs("interface.duplicatepeopletool", 350, 220)

        top.connect_signals(
            {
                "on_do_merge_clicked": self.__dummy,
                "on_help_show_clicked": self.__dummy,
                "on_delete_show_event": self.__dummy,
                "on_merge_ok_clicked": self.on_merge_ok_clicked,
                "destroy_passed_object": self.close,
                "on_help_clicked": self.on_help_clicked,
                "on_delete_merge_event": self.close,
                "on_delete_event": self.close,
            }
        )

        self.show()

    def build_menu_names(self, obj):
        return (_("Tool settings"), _("Find Duplicates tool"))

    def on_help_clicked(self, obj):
        """Display the relevant portion of Gramps manual"""

        display_help(WIKI_HELP_PAGE, WIKI_HELP_SEC)

    def ancestors_of(self, p1_id, id_list):
        if (not p1_id) or (p1_id in id_list):
            return
        id_list.append(p1_id)
        p1 = self.db.get_person_from_handle(p1_id)
        f1_id = p1.get_main_parents_family_handle()
        if f1_id:
            f1 = self.db.get_family_from_handle(f1_id)
            self.ancestors_of(f1.get_father_handle(), id_list)
            self.ancestors_of(f1.get_mother_handle(), id_list)

    def on_merge_ok_clicked(self, obj):
        threshold = self.menu.get_model()[self.menu.get_active()][1]
        self.use_soundex = int(self.soundex_obj.get_active())
        try:
            self.find_potentials(threshold)
        except AttributeError as msg:
            RunDatabaseRepair(str(msg), parent=self.window)
            return

        self.options.handler.options_dict["threshold"] = threshold
        self.options.handler.options_dict["soundex"] = self.use_soundex
        # Save options
        self.options.handler.save_options()

        if len(self.map) == 0:
            OkDialog(
                _("No matches found"),
                _("No potential duplicate people were found"),
                parent=self.window,
            )
        else:
            try:
                DuplicatePeopleToolMatches(
                    self.dbstate,
                    self.uistate,
                    self.track,
                    self.list,
                    self.map,
                    self.update,
                )
            except WindowActiveError:
                pass

    def find_potentials(self, thresh):
        self.progress = ProgressMeter(
            _("Find Duplicates"), _("Looking for duplicate people"), parent=self.window
        )

        index = 0
        males = {}
        females = {}
        self.map = {}

        length = self.db.get_number_of_people()

        self.progress.set_pass(_("Pass 1: Building preliminary lists"), length)

        for p1_id in self.db.iter_person_handles():
            self.progress.step()
            p1 = self.db.get_person_from_handle(p1_id)
            key = self.gen_key(get_surnames(p1.get_primary_name()))
            if p1.get_gender() == Person.MALE:
                if key in males:
                    males[key].append(p1_id)
                else:
                    males[key] = [p1_id]
            else:
                if key in females:
                    females[key].append(p1_id)
                else:
                    females[key] = [p1_id]

        self.progress.set_pass(_("Pass 2: Calculating potential matches"), length)

        for p1key in self.db.iter_person_handles():
            self.progress.step()
            p1 = self.db.get_person_from_handle(p1key)

            key = self.gen_key(get_surnames(p1.get_primary_name()))
            if p1.get_gender() == Person.MALE:
                remaining = males[key]
            else:
                remaining = females[key]

            # index = 0
            for p2key in remaining:
                # index += 1
                if p1key == p2key:
                    continue
                p2 = self.db.get_person_from_handle(p2key)
                if p2key in self.map:
                    (v, c) = self.map[p2key]
                    if v == p1key:
                        continue

                chance = self.compare_people(p1, p2)
                if chance >= thresh:
                    if p1key in self.map:
                        val = self.map[p1key]
                        if val[1] > chance:
                            self.map[p1key] = (p2key, chance)
                    else:
                        self.map[p1key] = (p2key, chance)

        self.list = sorted(self.map)
        self.length = len(self.list)
        self.progress.close()

    def gen_key(self, val):
        if self.use_soundex:
            try:
                return soundex(val)
            except UnicodeEncodeError:
                return val
        else:
            return val

    def compare_people(self, p1, p2):
        name1 = p1.get_primary_name()
        name2 = p2.get_primary_name()

        chance = self.name_match(name1, name2)
        if chance == -1:
            return -1

        birth1_ref = p1.get_birth_ref()
        if birth1_ref:
            birth1 = self.db.get_event_from_handle(birth1_ref.ref)
        else:
            birth1 = Event()

        death1_ref = p1.get_death_ref()
        if death1_ref:
            death1 = self.db.get_event_from_handle(death1_ref.ref)
        else:
            death1 = Event()

        birth2_ref = p2.get_birth_ref()
        if birth2_ref:
            birth2 = self.db.get_event_from_handle(birth2_ref.ref)
        else:
            birth2 = Event()

        death2_ref = p2.get_death_ref()
        if death2_ref:
            death2 = self.db.get_event_from_handle(death2_ref.ref)
        else:
            death2 = Event()

        value = self.date_match(birth1.get_date_object(), birth2.get_date_object())
        if value == -1:
            return -1
        chance += value

        value = self.date_match(death1.get_date_object(), death2.get_date_object())
        if value == -1:
            return -1
        chance += value

        value = self.place_match(birth1.get_place_handle(), birth2.get_place_handle())
        if value == -1:
            return -1
        chance += value

        value = self.place_match(death1.get_place_handle(), death2.get_place_handle())
        if value == -1:
            return -1
        chance += value

        ancestors = []
        self.ancestors_of(p1.get_handle(), ancestors)
        if p2.get_handle() in ancestors:
            return -1

        ancestors = []
        self.ancestors_of(p2.get_handle(), ancestors)
        if p1.get_handle() in ancestors:
            return -1

        f1_id = p1.get_main_parents_family_handle()
        f2_id = p2.get_main_parents_family_handle()

        if f1_id and f2_id:
            f1 = self.db.get_family_from_handle(f1_id)
            f2 = self.db.get_family_from_handle(f2_id)
            dad1_id = f1.get_father_handle()
            if dad1_id:
                dad1 = get_name_obj(self.db.get_person_from_handle(dad1_id))
            else:
                dad1 = None
            dad2_id = f2.get_father_handle()
            if dad2_id:
                dad2 = get_name_obj(self.db.get_person_from_handle(dad2_id))
            else:
                dad2 = None

            value = self.name_match(dad1, dad2)

            if value == -1:
                return -1

            chance += value

            mom1_id = f1.get_mother_handle()
            if mom1_id:
                mom1 = get_name_obj(self.db.get_person_from_handle(mom1_id))
            else:
                mom1 = None
            mom2_id = f2.get_mother_handle()
            if mom2_id:
                mom2 = get_name_obj(self.db.get_person_from_handle(mom2_id))
            else:
                mom2 = None

            value = self.name_match(mom1, mom2)
            if value == -1:
                return -1

            chance += value

        for f1_id in p1.get_family_handle_list():
            f1 = self.db.get_family_from_handle(f1_id)
            for f2_id in p2.get_family_handle_list():
                f2 = self.db.get_family_from_handle(f2_id)
                if p1.get_gender() == Person.FEMALE:
                    father1_id = f1.get_father_handle()
                    father2_id = f2.get_father_handle()
                    if father1_id and father2_id:
                        if father1_id == father2_id:
                            chance += 1
                        else:
                            father1 = self.db.get_person_from_handle(father1_id)
                            father2 = self.db.get_person_from_handle(father2_id)
                            fname1 = get_name_obj(father1)
                            fname2 = get_name_obj(father2)
                            value = self.name_match(fname1, fname2)
                            if value != -1:
                                chance += value
                else:
                    mother1_id = f1.get_mother_handle()
                    mother2_id = f2.get_mother_handle()
                    if mother1_id and mother2_id:
                        if mother1_id == mother2_id:
                            chance += 1
                        else:
                            mother1 = self.db.get_person_from_handle(mother1_id)
                            mother2 = self.db.get_person_from_handle(mother2_id)
                            mname1 = get_name_obj(mother1)
                            mname2 = get_name_obj(mother2)
                            value = self.name_match(mname1, mname2)
                            if value != -1:
                                chance += value
        return chance

    def name_compare(self, s1, s2):
        if self.use_soundex:
            try:
                return compare(s1, s2)
            except UnicodeEncodeError:
                return s1 == s2
        else:
            return s1 == s2

    def date_match(self, date1, date2):
        if date1.is_empty() or date2.is_empty():
            return 0
        if date1.is_equal(date2):
            return 1

        if date1.is_compound() or date2.is_compound():
            return self.range_compare(date1, date2)

        if date1.get_year() == date2.get_year():
            if date1.get_month() == date2.get_month():
                return 0.75
            if not date1.get_month_valid() or not date2.get_month_valid():
                return 0.75
            else:
                return -1
        else:
            return -1

    def range_compare(self, date1, date2):
        start_date_1 = date1.get_start_date()[0:3]
        start_date_2 = date2.get_start_date()[0:3]
        stop_date_1 = date1.get_stop_date()[0:3]
        stop_date_2 = date2.get_stop_date()[0:3]
        if date1.is_compound() and date2.is_compound():
            if (
                start_date_2 <= start_date_1 <= stop_date_2
                or start_date_1 <= start_date_2 <= stop_date_1
                or start_date_2 <= stop_date_1 <= stop_date_2
                or start_date_1 <= stop_date_2 <= stop_date_1
            ):
                return 0.5
            else:
                return -1
        elif date2.is_compound():
            if start_date_2 <= start_date_1 <= stop_date_2:
                return 0.5
            else:
                return -1
        else:
            if start_date_1 <= start_date_2 <= stop_date_1:
                return 0.5
            else:
                return -1

    def name_match(self, name, name1):
        if not name1 or not name:
            return 0

        srn1 = get_surnames(name)
        sfx1 = name.get_suffix()
        srn2 = get_surnames(name1)
        sfx2 = name1.get_suffix()

        if not self.name_compare(srn1, srn2):
            return -1
        if sfx1 != sfx2:
            if sfx1 != "" and sfx2 != "":
                return -1

        if name.get_first_name() == name1.get_first_name():
            return 1
        else:
            list1 = name.get_first_name().split()
            list2 = name1.get_first_name().split()

            if len(list1) < len(list2):
                return self.list_reduce(list1, list2)
            else:
                return self.list_reduce(list2, list1)

    def place_match(self, p1_id, p2_id):
        if p1_id == p2_id:
            return 1

        if not p1_id:
            name1 = ""
        else:
            p1 = self.db.get_place_from_handle(p1_id)
            name1 = p1.get_title()

        if not p2_id:
            name2 = ""
        else:
            p2 = self.db.get_place_from_handle(p2_id)
            name2 = p2.get_title()

        if not (name1 and name2):
            return 0
        if name1 == name2:
            return 1

        list1 = name1.replace(",", " ").split()
        list2 = name2.replace(",", " ").split()

        value = 0
        for name in list1:
            for name2 in list2:
                if name == name2:
                    value += 0.5
                elif name[0] == name2[0] and self.name_compare(name, name2):
                    value += 0.25
        return min(value, 1) if value else -1

    def list_reduce(self, list1, list2):
        value = 0
        for name in list1:
            for name2 in list2:
                if is_initial(name) and name[0] == name2[0]:
                    value += 0.25
                elif is_initial(name2) and name2[0] == name[0]:
                    value += 0.25
                elif name == name2:
                    value += 0.5
                elif name[0] == name2[0] and self.name_compare(name, name2):
                    value += 0.25
        return min(value, 1) if value else -1

    def __dummy(self, obj):
        """dummy callback, needed because a shared glade file is used for
        both toplevel windows and all signals must be handled.
        """
        pass


class DuplicatePeopleToolMatches(ManagedWindow):
    def __init__(self, dbstate, uistate, track, the_list, the_map, callback):
        ManagedWindow.__init__(self, uistate, track, self.__class__)

        self.dellist = set()
        self.list = the_list
        self.map = the_map
        self.length = len(self.list)
        self.update = callback
        self.db = dbstate.db
        self.dbstate = dbstate
        self.uistate = uistate

        top = Glade(toplevel="mergelist")
        window = top.toplevel
        self.set_window(window, top.get_object("title"), _("Potential Merges"))
        self.setup_configs("interface.duplicatepeopletoolmatches", 500, 350)

        self.mlist = top.get_object("mlist")
        top.connect_signals(
            {
                "destroy_passed_object": self.close,
                "on_do_merge_clicked": self.on_do_merge_clicked,
                "on_help_show_clicked": self.on_help_clicked,
                "on_delete_show_event": self.close,
                "on_merge_ok_clicked": self.__dummy,
                "on_help_clicked": self.__dummy,
                "on_delete_merge_event": self.__dummy,
                "on_delete_event": self.__dummy,
            }
        )
        self.db.connect("person-delete", self.person_delete)

        mtitles = [
            (_("Rating"), 3, 75),
            (_("First Person"), 1, 200),
            (_("Second Person"), 2, 200),
            ("", -1, 0),
        ]
        self.list = ListModel(self.mlist, mtitles, event_func=self.on_do_merge_clicked)

        self.redraw()
        self.show()

    def build_menu_names(self, obj):
        return (_("Merge candidates"), _("Merge persons"))

    def on_help_clicked(self, obj):
        """Display the relevant portion of Gramps manual"""

        display_help(WIKI_HELP_PAGE, WIKI_HELP_SEC)

    def redraw(self):
        list = []
        for p1key, p1data in self.map.items():
            if p1key in self.dellist:
                continue
            (p2key, c) = p1data
            if p2key in self.dellist:
                continue
            if p1key == p2key:
                continue
            list.append((c, p1key, p2key))

        self.list.clear()
        for c, p1key, p2key in list:
            c1 = "%5.2f" % c
            c2 = "%5.2f" % (100 - c)
            p1 = self.db.get_person_from_handle(p1key)
            p2 = self.db.get_person_from_handle(p2key)
            if not p1 or not p2:
                continue
            pn1 = name_displayer.display(p1)
            pn2 = name_displayer.display(p2)
            self.list.add([c1, pn1, pn2, c2], (p1key, p2key))

    def on_do_merge_clicked(self, obj):
        store, iter = self.list.selection.get_selected()
        if not iter:
            return

        (self.p1, self.p2) = self.list.get_object(iter)
        MergePerson(
            self.dbstate,
            self.uistate,
            self.track,
            self.p1,
            self.p2,
            self.on_update,
            True,
        )

    def on_update(self):
        if self.db.has_person_handle(self.p1):
            titanic = self.p2
        else:
            titanic = self.p1
        self.dellist.add(titanic)
        self.update()
        self.redraw()

    def update_and_destroy(self, obj):
        self.update(1)
        self.close()

    def person_delete(self, handle_list):
        """deal with person deletes outside of the tool"""
        self.dellist.update(handle_list)
        self.redraw()

    def __dummy(self, obj):
        """dummy callback, needed because a shared glade file is used for
        both toplevel windows and all signals must be handled.
        """
        pass


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
def name_of(p):
    if not p:
        return ""
    return "%s (%s)" % (name_displayer.display(p), p.get_handle())


def get_name_obj(person):
    if person:
        return person.get_primary_name()
    else:
        return None


def get_surnames(name):
    """Construct a full surname of the surnames"""
    return " ".join([surn.get_surname() for surn in name.get_surname_list()])


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class DuplicatePeopleToolOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)

        # Options specific for this report
        self.options_dict = {
            "soundex": 1,
            "threshold": 0.25,
        }
        self.options_help = {
            "soundex": (
                "=0/1",
                "Whether to use SoundEx codes",
                ["Do not use SoundEx", "Use SoundEx"],
                True,
            ),
            "threshold": ("=num", "Threshold for tolerance", "Floating point number"),
        }



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Tools/Database Processing/Sort Events"""

# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.sort import Sort
from gramps.gen.db import DbTxn
from gramps.gui.plug import MenuToolOptions, PluginWindows
from gramps.gen.plug.report import utils
from gramps.gen.plug.menu import (
    FilterOption,
    PersonOption,
    EnumeratedListOption,
    BooleanOption,
)


# ------------------------------------------------------------------------
#
# Private Functions
#
# ------------------------------------------------------------------------
def _get_sort_functions(sort):
    """
    Define the types of sorting that is available
    """
    return [
        (_("Date"), sort.by_date_key),
        (_("Type"), sort.by_event_type_key),
        (_("ID"), sort.by_event_id_key),
        (_("Description"), sort.by_event_description_key),
        (_("Place"), sort.by_event_place_key),
    ]


# -------------------------------------------------------------------------
#
# SortEvents
#
# -------------------------------------------------------------------------
class SortEvents(PluginWindows.ToolManagedWindowBatch):
    """
    This tool can be used to sort personal and family events by a variety
    of event attributes. It is primarily intended for re-organising data
    imported from an external source where the events are randomly ordered.
    """

    def get_title(self):
        """
        Return the window title.
        """
        return _("Sort Events")

    def initial_frame(self):
        """
        The name of the initial menu tab.
        """
        self.setup_configs("interface.sortevents", 500, 350)
        return _("Options")

    def run(self):
        """
        Perform the actual extraction of information.
        """
        menu = self.options.menu
        self.filter = menu.get_option_by_name("filter").get_filter()
        sort_func_num = menu.get_option_by_name("sort_by").get_value()
        self.sort_desc = menu.get_option_by_name("sort_desc").get_value()
        self.fam_events = menu.get_option_by_name("family_events").get_value()
        sort_functions = _get_sort_functions(Sort(self.db))
        self.sort_name = sort_functions[sort_func_num][0]
        self.sort_func = sort_functions[sort_func_num][1]
        self.sort = Sort(self.db)
        with DbTxn(_("Sort event changes"), self.db, batch=True) as trans:
            self.db.disable_signals()
            family_handles = self.sort_person_events(trans)
            if len(family_handles) > 0:
                self.sort_family_events(family_handles, trans)
        self.db.enable_signals()
        self.db.request_rebuild()

    def sort_person_events(self, trans):
        """
        Sort the personal events associated with the selected people.
        """
        people_handles = self.filter.apply(
            self.db, self.db.iter_person_handles(), user=self._user
        )
        self.progress.set_pass(
            _("Sorting personal events..."), self.db.get_number_of_people()
        )
        family_handles = []
        for handle in people_handles:
            person = self.db.get_person_from_handle(handle)
            self.progress.step()
            event_ref_list = person.get_event_ref_list()
            event_ref_list.sort(key=lambda x: self.sort_func(x.ref))
            if self.sort_desc:
                event_ref_list.reverse()
            if self.fam_events:
                family_handles.extend(person.get_family_handle_list())
            person.set_event_ref_list(event_ref_list)
            self.db.set_birth_death_index(person)
            self.db.commit_person(person, trans)
            self.change = True
        return family_handles

    def sort_family_events(self, family_handles, trans):
        """
        Sort the family events associated with the selected people.
        """
        self.progress.set_pass(_("Sorting family events..."), len(family_handles))
        for handle in family_handles:
            family = self.db.get_family_from_handle(handle)
            self.progress.step()
            event_ref_list = family.get_event_ref_list()
            event_ref_list.sort(key=lambda x: self.sort_func(x.ref))
            if self.sort_desc:
                event_ref_list.reverse()
            family.set_event_ref_list(event_ref_list)
            self.db.commit_family(family, trans)
            self.change = True


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class SortEventOptions(MenuToolOptions):
    """
    Define options and provides handling interface.
    """

    def __init__(self, name, person_id=None, dbstate=None):
        self.__db = dbstate.get_database()
        MenuToolOptions.__init__(self, name, person_id, dbstate)

    def add_menu_options(self, menu):
        """
        Define the options for the menu.
        """
        category_name = _("Tool Options")

        self.__filter = FilterOption(_("Filter"), 0)
        self.__filter.set_help(_("Select the people to sort"))
        menu.add_option(category_name, "filter", self.__filter)
        self.__filter.connect("value-changed", self.__filter_changed)

        self.__pid = PersonOption(_("Filter Person"))
        self.__pid.set_help(_("The center person for the filter"))
        menu.add_option(category_name, "pid", self.__pid)
        self.__pid.connect("value-changed", self.__update_filters)

        self.__update_filters()

        sort_by = EnumeratedListOption(_("Sort by"), 0)
        idx = 0
        for item in _get_sort_functions(Sort(self.__db)):
            sort_by.add_item(idx, item[0])
            idx += 1
        sort_by.set_help(_("Sorting method to use"))
        menu.add_option(category_name, "sort_by", sort_by)

        sort_desc = BooleanOption(_("Sort descending"), False)
        sort_desc.set_help(_("Set the sort order"))
        menu.add_option(category_name, "sort_desc", sort_desc)

        family_events = BooleanOption(_("Include family events"), True)
        family_events.set_help(_("Sort family events of the person"))
        menu.add_option(category_name, "family_events", family_events)

    def __filter_changed(self):
        """
        Handle filter change. If the filter is not specific to a person,
        disable the person option
        """
        filter_value = self.__filter.get_value()
        if filter_value == 0:  # "Entire Database" (as "include_single=False")
            self.__pid.set_available(False)
        else:
            # The other filters need a center person (assume custom ones too)
            self.__pid.set_available(True)

    def __update_filters(self):
        """
        Update the filter list based on the selected person
        """
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        filter_list = utils.get_person_filters(person, include_single=False)
        self.__filter.set_filters(filter_list)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2021-      Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# gui/columnorder.py

"""
Handle the csv format when exporting views
"""

# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------
import logging
import csv

# -------------------------------------------------------------------------
#
# GTK modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
#  -------------------------------------------------------------------------
from gramps.gen.config import config
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
__LOG = logging.getLogger(".CsvDialect")

CSV_DELIMITERS = {
    ",": ",",
    ";": ";",
    ":": ":",
    "|": "|",
    "\t": _("Tab", "character"),
}


class CsvDialect(Gtk.Box):
    """
    Use a custom format for csv export views
    """

    def __init__(self):
        """
        Used to set the csv dialect.
        We add the possibility to add a custom model where we
        can change the delimiter.
        """
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        hbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        hbox.set_spacing(10)
        hbox.pack_start(Gtk.Label(label=_("Choose your dialect")), False, False, 0)
        comment = Gtk.Label(
            label=_(
                "Changes are immediate.\n"
                "This is used when exporting views"
                " in CSV format."
            )
        )
        comment.set_justify(Gtk.Justification.LEFT)
        hbox.pack_start(comment, False, False, 0)
        hbox.pack_start(Gtk.Label(label=" "), False, True, 0)

        self.pack_start(hbox, True, True, 0)

        self.dialect = config.get("csv.dialect")
        self.delimiter = config.get("csv.delimiter")

        self.dialects = csv.list_dialects()
        self.dialects.append(_("Custom"))
        self.buttons = []
        button = None
        self.entry = Gtk.ComboBox()

        for dialect in self.dialects:
            title = dialect
            button = Gtk.RadioButton.new_with_mnemonic_from_widget(button, title)
            button.connect("toggled", self.on_toggled, self.entry)
            self.buttons.append(button)
            hbox.pack_start(button, False, True, 0)
            if dialect == self.dialect:
                button.set_active(True)

        lwidget = Gtk.Label(label=_("Delimiter:"))
        hbox.pack_start(lwidget, False, False, 0)

        store = Gtk.ListStore(str, str)
        default = 0
        for index, (key, value) in enumerate(CSV_DELIMITERS.items()):
            store.append((key, value))
            if key == self.delimiter:
                default = index
        self.entry.set_model(store)
        cell = Gtk.CellRendererText()
        self.entry.pack_start(cell, True)
        self.entry.add_attribute(cell, "text", 1)
        self.entry.set_active(default)
        self.entry.set_hexpand(False)
        self.entry.connect("changed", self.on_changed)
        hbox.pack_start(self.entry, False, True, 0)
        if self.dialect == _("Custom"):
            self.entry.set_sensitive(True)
        else:
            self.entry.set_sensitive(False)

    def on_changed(self, obj):
        """
        called when a button state change
        save is immediate
        """
        sep_iter = obj.get_active_iter()

        if sep_iter is not None:
            model = obj.get_model()
            sep = model[sep_iter][0]
            config.set("csv.delimiter", sep)

    def on_toggled(self, obj, entry):
        """
        called when a button state change
        save is immediate
        """
        if obj.get_active():
            button = obj.get_label()
            config.set("csv.dialect", button)
            if button == _("Custom"):
                entry.set_sensitive(True)
            else:
                entry.set_sensitive(False)
